#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os
import time

import ovs
from ovs.db.idl import Transaction, Idl, SchemaHelper

from libnmstate.plugin import NmstatePlugin
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge
from libnmstate.schema import OvsDB
from libnmstate.state import merge_dict
from libnmstate.error import NmstateNotImplementedError
from libnmstate.error import NmstateTimeoutError
from libnmstate.error import NmstatePermissionError
from libnmstate.error import NmstateValueError
from libnmstate.error import NmstatePluginError

TIMEOUT = 5

DEFAULT_OVS_DB_SOCKET_PATH = "/run/openvswitch/db.sock"
DEFAULT_OVS_SCHEMA_PATH = "/usr/share/openvswitch/vswitch.ovsschema"

NM_EXTERNAL_ID = "NM.connection.uuid"

GLOBAL_CONFIG_TABLE = "Open_vSwitch"

ALL_SUPPORTED_GLOBAL_COLUMNS = [OvsDB.EXTERNAL_IDS, OvsDB.OTHER_CONFIG]


class _Changes:
    def __init__(self, table_name, column_name, row_name, column_value):
        self.table_name = table_name
        self.column_name = column_name
        self.row_name = row_name
        self.column_value = column_value

    def __str__(self):
        return f"{self.__dict__}"


class NmstateOvsdbPlugin(NmstatePlugin):
    def __init__(self):
        self._schema = None
        self._idl = None
        self._transaction = None
        self._seq_no = 0
        self._load_schema()
        self._connect_to_ovs_db()

    @property
    def is_supplemental_only(self):
        return True

    def unload(self):
        if self._transaction:
            self._transaction.abort()
            self._transaction = None
        if self._idl:
            self._idl.close()
            self._idl = None

    def _load_schema(self):
        schema_path = os.environ.get(
            "OVS_SCHEMA_PATH", DEFAULT_OVS_SCHEMA_PATH
        )
        if not os.path.exists(schema_path):
            raise NmstateValueError(
                f"OVS schema file {schema_path} does not exist, "
                "please define the correct one via "
                "environment variable 'OVS_SCHEMA_PATH'"
            )
        if not os.access(schema_path, os.R_OK):
            raise NmstatePermissionError(
                f"Has no read permission to OVS schema file {schema_path}"
            )
        self._schema = SchemaHelper(schema_path)
        self._schema.register_columns(
            "Interface", [OvsDB.EXTERNAL_IDS, "name", "type"]
        )
        self._schema.register_columns("Bridge", [OvsDB.EXTERNAL_IDS, "name"])
        self._schema.register_columns(
            GLOBAL_CONFIG_TABLE,
            ALL_SUPPORTED_GLOBAL_COLUMNS,
        )

    def _connect_to_ovs_db(self):
        socket_path = os.environ.get(
            "OVS_DB_UNIX_SOCKET_PATH", DEFAULT_OVS_DB_SOCKET_PATH
        )
        if not os.path.exists(socket_path):
            raise NmstateValueError(
                f"OVS database socket file {socket_path} does not exist, "
                "please start the OVS daemon or define the socket path via "
                "environment variable 'OVS_DB_UNIX_SOCKET_PATH'"
            )
        if not os.access(socket_path, os.R_OK):
            raise NmstatePermissionError(
                f"Has no read permission to OVS db socket file {socket_path}"
            )

        self._idl = Idl(f"unix:{socket_path}", self._schema)
        self.refresh_content()
        if not self._idl.has_ever_connected():
            self._idl = None
            raise NmstatePluginError("Failed to connect to OVS DB")

    def refresh_content(self):
        if self._idl:
            timeout_end = time.time() + TIMEOUT
            self._idl.run()
            if self._idl.change_seqno == self._seq_no and self._seq_no:
                return
            while True:
                changed = self._idl.run()
                cur_seq_no = self._idl.change_seqno
                if cur_seq_no != self._seq_no or changed:
                    self._seq_no = cur_seq_no
                    return
                poller = ovs.poller.Poller()
                self._idl.wait(poller)
                poller.timer_wait(TIMEOUT * 1000)
                poller.block()
                if time.time() > timeout_end:
                    raise NmstateTimeoutError(
                        f"Plugin {self.name} timeout({TIMEOUT} "
                        "seconds) when refresh OVS database connection"
                    )

    @property
    def name(self):
        return "nmstate-plugin-ovsdb"

    @property
    def priority(self):
        return NmstatePlugin.DEFAULT_PRIORITY + 1

    @property
    def plugin_capabilities(self):
        return NmstatePlugin.PLUGIN_CAPABILITY_IFACE

    def get_interfaces(self):
        ifaces = []
        for row in self._idl.tables["Interface"].rows.values():
            if row.type in ("internal", "patch"):
                iface_type = InterfaceType.OVS_INTERFACE
            elif row.type == "system":
                # Let other plugin decide the interface type
                iface_type = InterfaceType.UNKNOWN
            else:
                continue

            ifaces.append(
                {
                    Interface.NAME: row.name,
                    Interface.TYPE: iface_type,
                    OvsDB.OVS_DB_SUBTREE: {
                        OvsDB.EXTERNAL_IDS: row.external_ids
                    },
                }
            )

        for row in self._idl.tables["Bridge"].rows.values():
            ifaces.append(
                {
                    Interface.NAME: row.name,
                    Interface.TYPE: InterfaceType.OVS_BRIDGE,
                    OvsDB.OVS_DB_SUBTREE: {
                        OvsDB.EXTERNAL_IDS: row.external_ids
                    },
                }
            )
        return ifaces

    def apply_changes(self, net_state, save_to_disk):
        # State might changed after other plugin invoked apply_changes()
        self.refresh_content()
        cur_iface_to_ext_ids = {}
        for iface_info in self.get_interfaces():
            cur_iface_to_ext_ids[iface_info[Interface.NAME]] = iface_info[
                OvsDB.OVS_DB_SUBTREE
            ][OvsDB.EXTERNAL_IDS]

        pending_changes = []
        for iface in net_state.ifaces.all_ifaces():
            if not iface.is_changed and not iface.is_desired:
                continue
            if not iface.is_up:
                continue
            if iface.type == OVSBridge.TYPE:
                table_name = "Bridge"
            elif OvsDB.OVS_DB_SUBTREE in iface.to_dict():
                table_name = "Interface"
            else:
                continue
            ids_after_nm_applied = cur_iface_to_ext_ids.get(iface.name, {})
            ids_before_nm_applied = (
                iface.to_dict()
                .get(OvsDB.OVS_DB_SUBTREE, {})
                .get(OvsDB.EXTERNAL_IDS, {})
            )
            original_desire_ids = iface.original_dict.get(
                OvsDB.OVS_DB_SUBTREE, {}
            ).get(OvsDB.EXTERNAL_IDS)

            desire_ids = []

            if original_desire_ids is None:
                desire_ids = ids_before_nm_applied
            else:
                desire_ids = original_desire_ids

            # should include external_id created by NetworkManager.
            if NM_EXTERNAL_ID in ids_after_nm_applied:
                desire_ids[NM_EXTERNAL_ID] = ids_after_nm_applied[
                    NM_EXTERNAL_ID
                ]
            if desire_ids != ids_after_nm_applied:
                pending_changes.append(
                    _generate_db_change_external_ids(
                        table_name, iface.name, desire_ids
                    )
                )
        global_config = net_state.desire_state.get(OvsDB.KEY)
        if global_config is not None:
            _merge_global_config(
                global_config, net_state.current_state.get(OvsDB.KEY, {})
            )
            pending_changes.extend(self._apply_global_config(global_config))

        if pending_changes:
            if not save_to_disk:
                raise NmstateNotImplementedError(
                    "ovsdb plugin does not support memory only changes"
                )
            elif self._idl:
                self._start_transaction()
                self._db_write(pending_changes)
                self._commit_transaction()

    def _apply_global_config(self, global_config):
        if global_config:
            return [
                _Changes(GLOBAL_CONFIG_TABLE, key, None, value)
                for key, value in global_config.items()
                if key in ALL_SUPPORTED_GLOBAL_COLUMNS
            ]
        else:
            # Remove all settings, assuming type is dict
            return [
                _Changes(GLOBAL_CONFIG_TABLE, key, None, {})
                for key in ALL_SUPPORTED_GLOBAL_COLUMNS
            ]

    def _db_write(self, changes):
        for change in changes:
            for row in self._idl.tables[change.table_name].rows.values():
                if (
                    change.table_name == GLOBAL_CONFIG_TABLE
                    or row.name == change.row_name
                ):
                    setattr(row, change.column_name, change.column_value)
                    break

    def _start_transaction(self):
        self._transaction = Transaction(self._idl)

    def _commit_transaction(self):
        if self._transaction:
            status = self._transaction.commit()
            timeout_end = time.time() + TIMEOUT
            while status == Transaction.INCOMPLETE:
                self._idl.run()
                poller = ovs.poller.Poller()
                self._idl.wait(poller)
                self._transaction.wait(poller)
                poller.timer_wait(TIMEOUT * 1000)
                poller.block()
                if time.time() > timeout_end:
                    raise NmstateTimeoutError(
                        f"Plugin {self.name} timeout({TIMEOUT} "
                        "seconds) when commit OVS database transaction"
                    )
                status = self._transaction.commit()

            if status == Transaction.SUCCESS:
                self.refresh_content()

            transaction_error = self._transaction.get_error()
            self._transaction = None

            if status not in (Transaction.SUCCESS, Transaction.UNCHANGED):
                raise NmstatePluginError(
                    f"Plugin {self.name} failure on commiting OVS database "
                    f"transaction: status: {status} "
                    f"error: {transaction_error}"
                )
        else:
            raise NmstatePluginError(
                "BUG: _commit_transaction() invoked with "
                "self._transaction is None"
            )

    def get_global_state(self):
        """
        Provide database information of global config table
        """
        info = {}
        for row in self._idl.tables[GLOBAL_CONFIG_TABLE].rows.values():
            for column_name in self._idl.tables[GLOBAL_CONFIG_TABLE].columns:
                info[column_name] = getattr(row, column_name)
            # There is only one row in global config table
            break

        return {OvsDB.KEY: info}


def _generate_db_change_external_ids(table_name, iface_name, desire_ids):
    if desire_ids and not isinstance(desire_ids, dict):
        raise NmstateValueError("Invalid external_ids, should be dictionary")

    # Convert all value to string
    for key, value in desire_ids.items():
        desire_ids[key] = str(value)

    return _Changes(table_name, OvsDB.EXTERNAL_IDS, iface_name, desire_ids)


def _merge_global_config(desire_global_config, current_global_config):
    """
    If any value been set to None, it means remove it.
    """
    # User can delete all global config by {OvsDB.KEY: {}}
    if desire_global_config != {}:
        merge_dict(desire_global_config, current_global_config)
        for column_name in ALL_SUPPORTED_GLOBAL_COLUMNS:
            if column_name in desire_global_config:
                keys_to_delete = [
                    key
                    for key, value in desire_global_config[column_name].items()
                    if value is None
                ]
                for key in keys_to_delete:
                    del desire_global_config[column_name][key]


NMSTATE_PLUGIN = NmstateOvsdbPlugin



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from contextlib import contextmanager
from copy import deepcopy

import time
import pytest
import yaml

import libnmstate
from libnmstate.error import NmstateKernelIntegerRoundedError
from libnmstate.error import NmstateLibnmError
from libnmstate.error import NmstateValueError
from libnmstate.error import NmstateVerificationError
from libnmstate.prettystate import PrettyState
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import LinuxBridge

from .testlib import assertlib
from .testlib import statelib
from .testlib.assertlib import assert_mac_address
from .testlib.bondlib import bond_interface
from .testlib.bridgelib import add_port_to_bridge
from .testlib.bridgelib import generate_vlan_filtering_config
from .testlib.bridgelib import generate_vlan_id_config
from .testlib.bridgelib import generate_vlan_id_range_config
from .testlib.bridgelib import linux_bridge
from .testlib.cmdlib import exec_cmd
from .testlib.env import is_fedora
from .testlib.env import is_ubuntu_kernel
from .testlib.ifacelib import get_mac_address
from .testlib.iproutelib import ip_monitor_assert_stable_link_up
from .testlib.statelib import show_only
from .testlib.vlan import vlan_interface


TEST_BRIDGE0 = "linux-br0"
TEST_TAP0 = "tap0"
ETH1 = "eth1"

BRIDGE_OPTIONS_YAML = """
options:
  group-forward-mask: 0
  mac-ageing-time: 300
  multicast-snooping: true
  stp:
    enabled: true
    forward-delay: 15
    hello-time: 2
    max-age: 20
    priority: 32768
"""

BRIDGE_PORT_YAML = """
stp-hairpin-mode: false
stp-path-cost: 100
stp-priority: 32
"""

VLAN_FILTER_PORT_YAML = """
    port:
    - name: eth1
      stp-hairpin-mode: false
      stp-path-cost: 100
      stp-priority: 32
      vlan:
        enable-native: true
        mode: trunk
        tag: 300
        trunk-tags:
        - id-range:
            max: 102
            min: 100
"""


@pytest.fixture
def bridge0_with_port0(port0_up):
    with _bridge0_with_port0(port0_up) as state:
        yield state


@pytest.fixture
def bridge0_with_port0_with_explicit_port_mac(port0_up):
    with _bridge0_with_port0(port0_up, use_port_mac=True) as state:
        yield state


@contextmanager
def _bridge0_with_port0(port0_up, use_port_mac=False):
    bridge_name = TEST_BRIDGE0
    port_name = port0_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port_name,))
    # Disable STP to avoid topology changes and the consequence link change.
    options_subtree = bridge_state[LinuxBridge.OPTIONS_SUBTREE]
    options_subtree[LinuxBridge.STP_SUBTREE][LinuxBridge.STP.ENABLED] = False

    extra_iface_state = None

    if use_port_mac:
        extra_iface_state = {Interface.MAC: get_mac_address(port_name)}

    with linux_bridge(
        bridge_name, bridge_state, extra_iface_state
    ) as desired_state:
        yield deepcopy(desired_state)


@pytest.fixture
def port0_vlan101(port0_up):
    vlan_id = 101
    vlan_base_iface = port0_up[Interface.KEY][0][Interface.NAME]
    port_name = "{}.{}".format(vlan_base_iface, vlan_id)
    with vlan_interface(port_name, vlan_id, vlan_base_iface):
        state = show_only((port_name,))
        yield state


@pytest.fixture
def bond0(port0_up):
    bond_name = "testbond0"
    port_name = port0_up[Interface.KEY][0][Interface.NAME]
    with bond_interface(bond_name, [port_name], create=False) as bond0:
        yield bond0


def test_create_and_remove_linux_bridge_with_min_desired_state():
    bridge_name = TEST_BRIDGE0
    with linux_bridge(bridge_name, bridge_subtree_state=None) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(bridge_name)


def test_create_and_remove_linux_bridge_with_one_port(port0_up):
    bridge_name = TEST_BRIDGE0
    port_name = port0_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port_name,))
    with linux_bridge(bridge_name, bridge_state) as desired_state:

        assertlib.assert_state(desired_state)

    assertlib.assert_absent(bridge_name)

    state = show_only((port_name,))
    assert state
    assert state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP


@pytest.mark.tier1
def test_create_and_remove_linux_bridge_with_two_ports(port0_up, port1_up):
    bridge_name = TEST_BRIDGE0
    port0_name = port0_up[Interface.KEY][0][Interface.NAME]
    port1_name = port1_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port0_name, port1_name))

    with linux_bridge(bridge_name, bridge_state) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(bridge_name)


@pytest.mark.tier1
def test_remove_bridge_and_keep_port_up(bridge0_with_port0, port0_up):
    bridge_name = bridge0_with_port0[Interface.KEY][0][Interface.NAME]
    port_name = port0_up[Interface.KEY][0][Interface.NAME]

    port_desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: port_name,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: bridge_name,
                Interface.STATE: InterfaceState.ABSENT,
            },
            port_desired_state[Interface.KEY][0],
        ]
    }

    libnmstate.apply(desired_state)

    current_state = show_only((bridge_name, port_name))

    assertlib.assert_state_match(port_desired_state)
    assert 1 == len(current_state[Interface.KEY])


@pytest.mark.tier1
def test_create_vlan_as_port_of_linux_bridge(port0_vlan101):
    bridge_name = TEST_BRIDGE0
    port_name = port0_vlan101[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port_name,))
    with linux_bridge(bridge_name, bridge_state) as desired_state:
        assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_create_vlan_over_linux_bridge(bridge0_with_port0):
    vlan_base_iface = TEST_BRIDGE0
    vlan_id = 101
    port_name = "{}.{}".format(vlan_base_iface, vlan_id)
    with vlan_interface(port_name, vlan_id, vlan_base_iface) as desired_state:
        assertlib.assert_state(desired_state)


@pytest.mark.tier1
@ip_monitor_assert_stable_link_up(TEST_BRIDGE0)
def test_add_port_to_existing_bridge(bridge0_with_port0, port1_up):
    desired_state = bridge0_with_port0
    bridge_iface_state = desired_state[Interface.KEY][0]
    bridge_state = bridge_iface_state[LinuxBridge.CONFIG_SUBTREE]
    port1_name = port1_up[Interface.KEY][0][Interface.NAME]
    _add_port_to_bridge(bridge_state, port1_name)

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
@pytest.mark.xfail(
    is_fedora(),
    reason=(
        "On Fedora 31+, users need to explicitly configure the port MAC "
        "due to changes to the default systemd config."
    ),
    raises=AssertionError,
    strict=True,
)
def test_linux_bridge_uses_the_port_mac_implicitly(
    port0_up, bridge0_with_port0
):
    port0_name = port0_up[Interface.KEY][0][Interface.NAME]
    current_state = show_only((TEST_BRIDGE0, port0_name))
    assert_mac_address(
        current_state, port0_up[Interface.KEY][0][Interface.MAC]
    )


@pytest.mark.tier1
def test_linux_bridge_uses_specified_mac_address(
    port0_up, bridge0_with_port0_with_explicit_port_mac
):
    port0_name = port0_up[Interface.KEY][0][Interface.NAME]
    current_state = show_only((TEST_BRIDGE0, port0_name))
    assert_mac_address(
        current_state, port0_up[Interface.KEY][0][Interface.MAC]
    )


def test_add_linux_bridge_with_empty_ipv6_static_address(port0_up):
    bridge_name = TEST_BRIDGE0
    port_name = port0_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port_name,))
    # Disable STP to avoid topology changes and the consequence link change.
    options_subtree = bridge_state[LinuxBridge.OPTIONS_SUBTREE]
    options_subtree[LinuxBridge.STP_SUBTREE][LinuxBridge.STP.ENABLED] = False

    extra_iface_state = {
        Interface.IPV6: {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.AUTOCONF: False,
            InterfaceIPv6.DHCP: False,
        }
    }
    with linux_bridge(
        bridge_name, bridge_state, extra_iface_state
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(bridge_name)


def test_add_linux_bridge_with_empty_ipv6_static_address_with_stp(port0_up):
    bridge_name = TEST_BRIDGE0
    port_name = port0_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port_name,))
    options_subtree = bridge_state[LinuxBridge.OPTIONS_SUBTREE]
    options_subtree[LinuxBridge.STP_SUBTREE][LinuxBridge.STP.ENABLED] = True

    extra_iface_state = {
        Interface.IPV6: {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.AUTOCONF: False,
            InterfaceIPv6.DHCP: False,
        }
    }
    with linux_bridge(
        bridge_name, bridge_state, extra_iface_state
    ) as desired_state:
        assertlib.assert_state(desired_state)

    assertlib.assert_absent(bridge_name)


def test_linux_bridge_add_port_with_name_only(bridge0_with_port0, port1_up):
    desired_state = bridge0_with_port0
    bridge_iface_state = desired_state[Interface.KEY][0]
    bridge_state = bridge_iface_state[LinuxBridge.CONFIG_SUBTREE]
    port1_name = port1_up[Interface.KEY][0][Interface.NAME]
    bridge_state[LinuxBridge.PORT_SUBTREE].append(
        {LinuxBridge.Port.NAME: port1_name}
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state_match(desired_state)


@pytest.mark.tier1
def test_replace_port_on_linux_bridge(port0_vlan101, port1_up):
    bridge_name = TEST_BRIDGE0
    vlan_port0_name = port0_vlan101[Interface.KEY][0][Interface.NAME]
    port1_name = port1_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((vlan_port0_name,))
    with linux_bridge(bridge_name, bridge_state) as state:
        brconf_state = state[Interface.KEY][0][LinuxBridge.CONFIG_SUBTREE]
        brconf_state[LinuxBridge.PORT_SUBTREE] = [
            {LinuxBridge.Port.NAME: port1_name}
        ]
        libnmstate.apply(state)

        br_state = show_only((bridge_name,))
        brconf_state = br_state[Interface.KEY][0][LinuxBridge.CONFIG_SUBTREE]
        br_ports_state = brconf_state[LinuxBridge.PORT_SUBTREE]
        assert 1 == len(br_ports_state)
        assert port1_name == br_ports_state[0][LinuxBridge.Port.NAME]

        port_state = show_only((vlan_port0_name,))
        assert (
            InterfaceState.UP == port_state[Interface.KEY][0][Interface.STATE]
        )


def test_linux_bridge_over_bond_over_port_in_one_transaction(bond0):
    bridge_name = TEST_BRIDGE0
    bond_name = bond0[Interface.KEY][0][Interface.NAME]
    bridge_config_state = _create_bridge_subtree_config((bond_name,))
    with linux_bridge(
        bridge_name, bridge_config_state, create=False
    ) as bridge0:
        desired_state = bond0
        bridge_state = bridge0[Interface.KEY][0]
        desired_state[Interface.KEY].append(bridge_state)
        libnmstate.apply(desired_state)

        assertlib.assert_state_match(desired_state)

    assertlib.assert_absent(bridge_name)


def test_rollback_for_linux_bridge():
    original_state = libnmstate.show()
    bridge_name = TEST_BRIDGE0
    bridge_state = _create_bridge_subtree_config(())
    with pytest.raises(NmstateVerificationError):
        with linux_bridge(bridge_name, bridge_state) as desired_state:
            desired_state[Interface.KEY][0]["invalid_key"] = "foo"
            libnmstate.apply(desired_state)

    time.sleep(5)  # Give some time for NetworkManager to rollback
    current_state = libnmstate.show()
    assert original_state == current_state


@pytest.mark.tier1
def test_activate_empty_bridge_does_not_blocked_by_dhcp():
    bridge_name = TEST_BRIDGE0
    bridge_state = None
    extra_iface_state = {
        Interface.IPV4: {
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.DHCP: True,
        },
        Interface.IPV6: {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.AUTOCONF: True,
            InterfaceIPv6.DHCP: True,
        },
    }
    with linux_bridge(
        bridge_name, bridge_state, extra_iface_state
    ) as desired_state:
        assertlib.assert_state(desired_state)


class TestVlanFiltering:
    ACCESS_TAG = 300

    @pytest.fixture
    def bridge_with_trunk_port_and_native_config(self, port0_up):
        trunk_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.TRUNK,
            native_vlan=True,
            tag=TestVlanFiltering.ACCESS_TAG,
            trunk_tags=generate_vlan_id_config(100, 101, 102),
        )
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(trunk_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as state:
            yield state

    @pytest.fixture
    def bridge_with_access_port_config(self, port0_up):
        access_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.ACCESS,
            tag=TestVlanFiltering.ACCESS_TAG,
        )
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(access_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as state:
            yield state

    @pytest.mark.parametrize(
        "is_native_vlan", [True, False], ids=["native", "not-native"]
    )
    def test_trunk_port_config(self, is_native_vlan, port0_up):
        trunk_tags = generate_vlan_id_config(100, 101)
        trunk_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.TRUNK,
            tag=TestVlanFiltering.ACCESS_TAG if is_native_vlan else None,
            trunk_tags=trunk_tags,
            native_vlan=is_native_vlan,
        )

        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(trunk_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    @pytest.mark.parametrize(
        "is_native_vlan", [True, False], ids=["native", "not-native"]
    )
    def test_port_with_multiple_trunk_ports_config(
        self, is_native_vlan, port0_up
    ):
        trunk_tags = generate_vlan_id_config(100, 102)
        trunk_tags.append(generate_vlan_id_range_config(200, 299))
        trunk_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.TRUNK,
            tag=TestVlanFiltering.ACCESS_TAG if is_native_vlan else None,
            trunk_tags=trunk_tags,
            native_vlan=is_native_vlan,
        )

        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(trunk_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_access_port_config(self, port0_up):
        access_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.ACCESS, tag=TestVlanFiltering.ACCESS_TAG
        )
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(access_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_update_trunk_port_to_access_port(
        self, bridge_with_trunk_port_and_native_config, port0_up
    ):
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))

        new_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.ACCESS,
            tag=TestVlanFiltering.ACCESS_TAG + 5,
        )
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(new_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_update_trunk_port_tag_ids(
        self, bridge_with_trunk_port_and_native_config, port0_up
    ):
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))

        new_trunk_tags = generate_vlan_id_config(600, 602)
        new_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.TRUNK, trunk_tags=new_trunk_tags
        )
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(new_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_update_access_port_tag_id(
        self, bridge_with_access_port_config, port0_up
    ):
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))

        new_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.ACCESS,
            tag=TestVlanFiltering.ACCESS_TAG * 2,
        )
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(new_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_activate_vlan_filtering_on_bridge(
        self, bridge0_with_port0, port0_up
    ):
        port_name = port0_up[Interface.KEY][0][Interface.NAME]
        bridge_state = _create_bridge_subtree_config((port_name,))

        new_trunk_tags = generate_vlan_id_config(600, 602)
        new_port_state = generate_vlan_filtering_config(
            LinuxBridge.Port.Vlan.Mode.TRUNK, trunk_tags=new_trunk_tags
        )
        bridge_state[LinuxBridge.PORT_SUBTREE][0].update(new_port_state)
        with linux_bridge(TEST_BRIDGE0, bridge_state) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_disable_vlan_filtering_on_bridge(
        self,
        bridge_with_trunk_port_and_native_config,
    ):
        bridge_state = bridge_with_trunk_port_and_native_config[Interface.KEY][
            0
        ]
        bridge_config_subtree = bridge_state[LinuxBridge.CONFIG_SUBTREE]
        bridge_ports = bridge_config_subtree[LinuxBridge.PORT_SUBTREE]
        bridge_ports[0].update({LinuxBridge.Port.VLAN_SUBTREE: {}})

        with linux_bridge(
            TEST_BRIDGE0, bridge_config_subtree
        ) as desired_state:
            assertlib.assert_state_match(desired_state)

    def test_pretty_state_port_name_first(
        self, bridge_with_trunk_port_and_native_config
    ):
        current_state = show_only((TEST_BRIDGE0,))
        pretty_state = PrettyState(current_state)
        assert VLAN_FILTER_PORT_YAML in pretty_state.yaml

    def test_show_saved_config_with_vlan_filter_bridge_down(
        self,
        bridge_with_trunk_port_and_native_config,
    ):
        running_state = statelib.show_only((TEST_BRIDGE0,))
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: TEST_BRIDGE0,
                        Interface.STATE: InterfaceState.DOWN,
                    }
                ]
            }
        )
        saved_state = statelib.show_saved_config_only((TEST_BRIDGE0,))

        assert (
            saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
        )
        assertlib.assert_state_match_full(saved_state, running_state)


@pytest.fixture
def bridge_unmanaged_port():
    bridge_config = _create_bridge_subtree_config([])
    with linux_bridge(TEST_BRIDGE0, bridge_config):
        with dummy0_as_port(TEST_BRIDGE0):
            current_state = show_only((TEST_BRIDGE0,))
            yield current_state


def test_bridge_with_unmanaged_ports(bridge_unmanaged_port):
    bridge_state = bridge_unmanaged_port[Interface.KEY][0]
    port_subtree = bridge_state[LinuxBridge.CONFIG_SUBTREE][
        LinuxBridge.PORT_SUBTREE
    ]

    assert port_subtree[0][LinuxBridge.Port.NAME] == "dummy0"


@contextmanager
def dummy0_as_port(controller):
    exec_cmd(("ip", "link", "add", "dummy0", "type", "dummy"), check=True)
    try:
        exec_cmd(("ip", "link", "set", "dummy0", "up"), check=True)
        exec_cmd(
            ("nmcli", "dev", "set", "dummy0", "managed", "no"), check=True
        )
        exec_cmd(
            ("ip", "link", "set", "dummy0", "master", controller), check=True
        )
        yield
    finally:
        exec_cmd(("ip", "link", "delete", "dummy0"))
        exec_cmd(("nmcli", "c", "del", "dummy0"))


def test_add_invalid_port_ip_config(eth1_up):
    desired_state = eth1_up
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.ENABLED] = True
    desired_state[Interface.KEY][0][Interface.IPV4][InterfaceIP.DHCP] = True
    bridge_state = _create_bridge_subtree_config(("eth1",))
    with pytest.raises(NmstateValueError):
        with linux_bridge(
            TEST_BRIDGE0, bridge_state, create=False
        ) as lb_state:
            desired_state[Interface.KEY].append(lb_state[Interface.KEY][0])
            libnmstate.apply(desired_state)


def _add_port_to_bridge(bridge_state, ifname):
    port_state = yaml.load(BRIDGE_PORT_YAML, Loader=yaml.SafeLoader)
    add_port_to_bridge(bridge_state, ifname, port_state)


def _create_bridge_subtree_config(port_names):
    bridge_state = yaml.load(BRIDGE_OPTIONS_YAML, Loader=yaml.SafeLoader)

    for port in port_names:
        port_state = yaml.load(BRIDGE_PORT_YAML, Loader=yaml.SafeLoader)
        add_port_to_bridge(bridge_state, port, port_state)

    return bridge_state


@pytest.mark.tier1
def test_change_linux_bridge_group_addr(bridge0_with_port0):
    iface_state = bridge0_with_port0[Interface.KEY][0]
    iface_state[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.OPTIONS_SUBTREE][
        LinuxBridge.Options.GROUP_ADDR
    ] = "01:80:C2:00:00:04"

    desired_state = {Interface.KEY: [iface_state]}

    libnmstate.apply(desired_state)

    assertlib.assert_state_match(desired_state)


@pytest.mark.skipif(
    not is_ubuntu_kernel(),
    reason="Only 250 HZ kernel will fail on NmstateKernelIntergerRounded "
    "for linux bridge MULTICAST_STARTUP_QUERY_INTERVAL option",
)
def test_linux_bridge_option_integer_rounded_on_ubuntu_kernel(
    bridge0_with_port0,
):
    iface_state = bridge0_with_port0[Interface.KEY][0]
    iface_state[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.OPTIONS_SUBTREE][
        LinuxBridge.Options.MULTICAST_STARTUP_QUERY_INTERVAL
    ] = 3125

    desired_state = {Interface.KEY: [iface_state]}

    with pytest.raises(NmstateKernelIntegerRoundedError):
        libnmstate.apply(desired_state)


@pytest.mark.tier1
def test_moving_ports_from_absent_interface(bridge0_with_port0):
    iface_state = bridge0_with_port0[Interface.KEY][0]
    iface_state[Interface.NAME] = "linux-br1"

    libnmstate.apply(
        {
            Interface.KEY: [
                iface_state,
                {
                    Interface.NAME: TEST_BRIDGE0,
                    Interface.STATE: InterfaceState.ABSENT,
                },
            ]
        }
    )

    assertlib.assert_state_match({Interface.KEY: [iface_state]})
    assertlib.assert_absent(TEST_BRIDGE0)

    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "linux-br1",
                    Interface.STATE: InterfaceState.ABSENT,
                },
            ]
        }
    )


@pytest.mark.xfail(
    reason="https://bugzilla.redhat.com/1869063 and "
    "https://bugzilla.redhat.com/1869079",
    raises=(NmstateLibnmError, NmstateVerificationError),
    strict=True,
)
def test_linux_bridge_replace_unmanaged_port(bridge_unmanaged_port, eth1_up):
    iface_state = bridge_unmanaged_port[Interface.KEY][0]
    iface_state[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.PORT_SUBTREE] = [
        {LinuxBridge.Port.NAME: "eth1"}
    ]
    libnmstate.apply({Interface.KEY: [iface_state]})


@pytest.fixture
def bridge0_with_tap_port():
    exec_cmd(f"ip tuntap add name {TEST_TAP0} mode tap".split(), check=True)
    exec_cmd(f"ip link add {TEST_BRIDGE0} type bridge".split(), check=True)
    exec_cmd(
        f"ip link set {TEST_TAP0} master {TEST_BRIDGE0}".split(), check=True
    )
    exec_cmd(f"ip link set {TEST_TAP0} up".split(), check=True)
    exec_cmd(f"ip link set {TEST_BRIDGE0} up".split(), check=True)
    yield
    exec_cmd(f"ip link del {TEST_TAP0}".split())
    exec_cmd(f"ip link del {TEST_BRIDGE0}".split())
    exec_cmd(f"nmcli c del {TEST_TAP0}".split())
    exec_cmd(f"nmcli c del {TEST_BRIDGE0}".split())


def test_ignore_unmanged_tap_as_bridge_port(bridge0_with_tap_port, port0_up):
    """
    The unknown interface should be still bridge port along with other ports
    """
    with _bridge0_with_port0(port0_up) as state:
        state[Interface.KEY][0][LinuxBridge.CONFIG_SUBTREE][
            LinuxBridge.PORT_SUBTREE
        ].append({LinuxBridge.Port.NAME: TEST_TAP0})
        assertlib.assert_state_match(state)


def test_explicitly_ignore_a_bridge_port(bridge0_with_port0, port1_up):
    """
    The ignored interface should be still bridge port along with other ports
    """
    bridge_info = bridge0_with_port0[Interface.KEY][0]
    port0_name = bridge_info[LinuxBridge.CONFIG_SUBTREE][
        LinuxBridge.PORT_SUBTREE
    ][0][LinuxBridge.Port.NAME]
    port1_name = port1_up[Interface.KEY][0][Interface.NAME]
    _add_port_to_bridge(bridge_info[LinuxBridge.CONFIG_SUBTREE], port1_name)
    desired_state = {
        Interface.KEY: [
            bridge_info,
            {
                Interface.NAME: port0_name,
                Interface.STATE: InterfaceState.IGNORE,
            },
        ]
    }
    print(desired_state)
    libnmstate.apply(desired_state)
    state = show_only((TEST_BRIDGE0,))
    cur_port_config = state[Interface.KEY][0][LinuxBridge.CONFIG_SUBTREE][
        LinuxBridge.PORT_SUBTREE
    ]
    assert cur_port_config[0][LinuxBridge.Port.NAME] == port0_name
    assert cur_port_config[1][LinuxBridge.Port.NAME] == port1_name


@pytest.mark.xfail(
    reason="https://bugzilla.redhat.com/1855160",
    raises=NmstateVerificationError,
)
def test_change_multicast_snooping_from_false_to_true(port0_up, port1_up):
    bridge_name = TEST_BRIDGE0
    port0_name = port0_up[Interface.KEY][0][Interface.NAME]
    port1_name = port1_up[Interface.KEY][0][Interface.NAME]
    bridge_state = _create_bridge_subtree_config((port0_name, port1_name))
    bridge_state[LinuxBridge.OPTIONS_SUBTREE][
        LinuxBridge.Options.MULTICAST_SNOOPING
    ] = False
    with linux_bridge(bridge_name, bridge_state, create=True) as lb_state:
        bridge = lb_state[Interface.KEY][0][LinuxBridge.CONFIG_SUBTREE]
        bridge[LinuxBridge.OPTIONS_SUBTREE][
            LinuxBridge.Options.MULTICAST_SNOOPING
        ] = True
        libnmstate.apply(lb_state)


def test_empty_state_does_not_change_bridge_options(bridge0_with_port0):
    original_state = show_only((TEST_BRIDGE0,))
    libnmstate.apply({Interface.KEY: [{Interface.NAME: TEST_BRIDGE0}]})
    assertlib.assert_state_match(original_state)


def test_create_bridge_with_mixed_case_group_addr():
    bridge_state = _create_bridge_subtree_config([])

    bridge_state[LinuxBridge.OPTIONS_SUBTREE][
        LinuxBridge.Options.GROUP_ADDR
    ] = "01:80:C2:00:00:0a"
    print(bridge_state)
    with linux_bridge(TEST_BRIDGE0, bridge_state) as lb_state:
        assertlib.assert_state_match(lb_state)


def test_show_saved_config_with_bridge_down(bridge0_with_port0):
    running_state = statelib.show_only((TEST_BRIDGE0,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: TEST_BRIDGE0,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((TEST_BRIDGE0,))

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os

from . import error
from . import schema

from .netapplier import apply
from .netapplier import commit
from .netapplier import rollback
from .netinfo import show
from .netinfo import show_running_config
from .netinfo import show_saved_config

from .prettystate import PrettyState


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

__all__ = [
    "PrettyState",
    "apply",
    "commit",
    "error",
    "rollback",
    "schema",
    "show",
    "show_running_config",
    "show_saved_config",
]


def _get_version():
    with open(os.path.join(ROOT_DIR, "VERSION")) as f:
        version = f.read().strip()
    return version


__version__ = _get_version()



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import Bond
from libnmstate.schema import BondMode

from ..testlib import assertlib
from ..testlib import cmdlib
from ..testlib.bondlib import bond_interface


BOND0 = "bondtest0"


def test_bond_all_zero_ad_actor_system_been_ignored():
    extra_iface_state = {
        Bond.CONFIG_SUBTREE: {
            Bond.MODE: BondMode.LACP,
            Bond.OPTIONS_SUBTREE: {"ad_actor_system": "00:00:00:00:00:00"},
        }
    }
    with bond_interface(
        name=BOND0, port=[], extra_iface_state=extra_iface_state, create=True
    ):
        _, output, _ = cmdlib.exec_cmd(
            f"nmcli --fields bond.options c show {BOND0}".split(), check=True
        )
        assert "ad_actor_system" not in output

    assertlib.assert_absent(BOND0)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
from operator import attrgetter

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType


DEFAULT_MAC_ADDRESS = "00:00:00:00:00:00"


class NisporPluginBaseIface:
    def __init__(self, np_iface):
        self._np_iface = np_iface

    @property
    def np_iface(self):
        return self._np_iface

    @property
    def mac(self):
        if self._np_iface.mac_address:
            return self._np_iface.mac_address.upper()
        return DEFAULT_MAC_ADDRESS

    @property
    def mtu(self):
        return self._np_iface.mtu

    @property
    def type(self):
        return InterfaceType.UNKNOWN

    @property
    def state(self):
        np_state = self._np_iface.state
        np_flags = self._np_iface.flags
        if np_state == "up" or "up" in np_flags or "running" in np_flags:
            return InterfaceState.UP
        elif np_state == "down":
            return InterfaceState.DOWN
        else:
            logging.debug(
                f"Got unexpect nispor interface state {np_state} for "
                f"{self._np_iface}"
            )
            return InterfaceState.DOWN

    def _ip_info(self, config_only):
        return {
            Interface.IPV4: NisporPlugintIpState(
                Interface.IPV4, self.np_iface.ipv4
            ).to_dict(config_only),
            Interface.IPV6: NisporPlugintIpState(
                Interface.IPV6, self.np_iface.ipv6
            ).to_dict(config_only),
        }

    def to_dict(self, config_only):
        iface_info = {
            Interface.NAME: self.np_iface.name,
            Interface.TYPE: self.type,
            Interface.STATE: self.state,
            Interface.MAC: self.mac,
        }
        if self.mtu:
            iface_info[Interface.MTU] = self.mtu
        ip_info = self._ip_info(config_only)
        if ip_info:
            iface_info.update(ip_info)

        return iface_info


class NisporPlugintIpState:
    def __init__(self, family, np_ip_state):
        self._family = family
        self._np_ip_state = np_ip_state
        self._addresses = []
        if np_ip_state:
            self._addresses = sorted(
                np_ip_state.addresses, key=attrgetter("address")
            )

    @property
    def _is_ipv6(self):
        return self._family == Interface.IPV6

    def _has_dhcp_address(self):
        return any(
            _is_dhcp_addr(addr, self._is_ipv6) for addr in self._addresses
        )

    def _has_autoconf_address(self):
        return self._is_ipv6 and any(
            _is_autoconf_addr(addr) for addr in self._addresses
        )

    def to_dict(self, config_only):
        if not self._addresses or not self._np_ip_state:
            return {InterfaceIP.ENABLED: False, InterfaceIP.ADDRESS: []}
        else:
            if config_only:
                addresses = [
                    addr
                    for addr in self._addresses
                    if not _is_autoconf_addr(addr)
                    and not _is_dhcp_addr(addr, self._is_ipv6)
                ]
            else:
                addresses = self._addresses
            info = {
                InterfaceIP.ENABLED: True,
                InterfaceIP.ADDRESS: [
                    {
                        InterfaceIP.ADDRESS_IP: addr.address,
                        InterfaceIP.ADDRESS_PREFIX_LENGTH: addr.prefix_len,
                    }
                    for addr in addresses
                ],
            }
            if self._has_dhcp_address():
                info[InterfaceIP.DHCP] = True
            if self._has_autoconf_address():
                info[InterfaceIPv6.AUTOCONF] = True
            return info


def _is_dhcp_addr(np_addr, is_ipv6):
    if is_ipv6:
        return np_addr.valid_lft != "forever" and np_addr.prefix_len == 128
    else:
        return np_addr.valid_lft != "forever"


def _is_autoconf_addr(np_addr):
    return np_addr.valid_lft != "forever" and np_addr.prefix_len == 64



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import InterfaceIP

FOO_IFACE_NAME = "foo"
MAC_ADDRESS1 = "01:23:45:67:89:AB"

IPV6_ADDRESS1 = "2001:db8:1::1"
IPV6_ADDRESS1_FULL = "2001:db8:1:0:0:0:0:1"
IPV6_ADDRESS2 = "2001:db8:2::1"
IPV6_LINK_LOCAL_ADDRESS1 = "fe80::1"
IPV6_ADDRESSES = [
    {
        InterfaceIP.ADDRESS_IP: IPV6_ADDRESS1,
        InterfaceIP.ADDRESS_PREFIX_LENGTH: 64,
    },
    {
        InterfaceIP.ADDRESS_IP: IPV6_ADDRESS2,
        InterfaceIP.ADDRESS_PREFIX_LENGTH: 64,
    },
]

IPV4_ADDRESS1 = "192.0.2.251"
IPV4_ADDRESS2 = "192.0.2.252"
IPV4_ADDRESSES = [
    {
        InterfaceIP.ADDRESS_IP: IPV4_ADDRESS1,
        InterfaceIP.ADDRESS_PREFIX_LENGTH: 24,
    },
    {
        InterfaceIP.ADDRESS_IP: IPV4_ADDRESS2,
        InterfaceIP.ADDRESS_PREFIX_LENGTH: 24,
    },
]

PORT1_IFACE_NAME = "port1"
PORT2_IFACE_NAME = "port2"



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import RouteRule
from libnmstate.schema import Route

from libnmstate.ifaces import BaseIface
from libnmstate.route import RouteState
from libnmstate.route_rule import RouteRuleEntry
from libnmstate.route_rule import RouteRuleState

from .testlib.ifacelib import gen_two_static_ip_ifaces
from .testlib.routelib import IPV4_ROUTE_IFACE_NAME
from .testlib.routelib import IPV4_ROUTE_TABLE_ID
from .testlib.routelib import IPV6_ROUTE_IFACE_NAME
from .testlib.routelib import IPV6_ROUTE_TABLE_ID
from .testlib.routelib import gen_ipv4_route
from .testlib.routelib import gen_ipv6_route


IPV4_ROUTE_RULE_FROM = "198.51.100.0/24"
IPV4_ROUTE_RULE_TO = "198.0.2.0/24"
IPV6_ROUTE_RULE_FROM = "2001:db8:a::/64"
IPV6_ROUTE_RULE_TO = "2001:db8:1::/64"


class TestRouteRuleEntry:
    def test_hash_unique(self):
        rule = _create_route_rule("198.51.100.0/24", "192.0.2.1", 50, 103)
        assert hash(rule) == hash(rule)

    def test_obj_unique(self):
        rule0 = _create_route_rule("198.51.100.0/24", "192.0.2.1", 50, 103)
        rule1 = _create_route_rule("2001:db8:a::/64", "2001:db8:1::a", 51, 104)
        rule0_clone = _create_route_rule(
            "198.51.100.0/24", "192.0.2.1", 50, 103
        )
        assert rule0 == rule0_clone
        assert rule0 != rule1

    def test_obj_unique_without_table(self):
        rule_with_default_table_id = _create_route_rule(
            "198.51.100.0/24",
            "192.0.2.1",
            103,
            RouteRule.USE_DEFAULT_ROUTE_TABLE,
        )

        rule_without_table_id = _create_route_rule(
            "198.51.100.0/24", "192.0.2.1", 103, None
        )

        assert rule_without_table_id == rule_with_default_table_id

    def test_obj_unique_without_priority(self):
        rule_with_default_priority = _create_route_rule(
            "198.51.100.0/24", "192.0.2.1", RouteRule.USE_DEFAULT_PRIORITY, 50
        )

        rule_without_priority = _create_route_rule(
            "198.51.100.0/24", "192.0.2.1", None, 50
        )

        assert rule_without_priority == rule_with_default_priority

    def test_normal_object_as_dict(self):
        rule = _create_route_rule_dict(
            "198.51.100.0/24", "192.0.2.1/32", 50, 103
        )
        rule_obj = RouteRuleEntry(rule)
        assert rule_obj.to_dict() == rule

    @pytest.mark.parametrize(
        "ip_ver_addrs",
        [
            (
                "198.51.100.0",
                "198.51.100.0/32",
                "192.0.2.0/24",
                "192.0.2.0/24",
            ),
            (
                "2001:db8:a::1",
                "2001:db8:a::1/128",
                "2001:db8:b::/64",
                "2001:db8:b::/64",
            ),
            (
                "192.0.2.0/24",
                "192.0.2.0/24",
                "198.51.100.0",
                "198.51.100.0/32",
            ),
            (
                "2001:db8:b::/64",
                "2001:db8:b::/64",
                "2001:db8:a::1",
                "2001:db8:a::1/128",
            ),
        ],
        ids=["ipv4_from", "ipv6_from", "ipv4_to", "ipv6_to"],
    )
    def test_host_only_ip_address(self, ip_ver_addrs):
        ip_from, expected_ip_from, ip_to, expected_ip_to = ip_ver_addrs
        rule = _create_route_rule_dict(ip_from, ip_to, 50, 103)
        expected_rule = _create_route_rule_dict(
            expected_ip_from, expected_ip_to, 50, 103
        )
        rule_obj = RouteRuleEntry(rule)
        assert rule_obj.to_dict() == expected_rule

    @pytest.mark.parametrize(
        "ip_ver_addrs",
        [
            (
                "198.51.100.1/24",
                "198.51.100.0/24",
            ),
            (
                "2001:db8:a::1/64",
                "2001:db8:a::/64",
            ),
        ],
        ids=["ipv4", "ipv6"],
    )
    def test_ip_network_address_with_host_bits(self, ip_ver_addrs):
        ip_addr, expected_ip_addr = ip_ver_addrs
        rule = _create_route_rule_dict(ip_addr, ip_addr, 50, 103)
        expected_rule = _create_route_rule_dict(
            expected_ip_addr, expected_ip_addr, 50, 103
        )
        rule_obj = RouteRuleEntry(rule)
        assert rule_obj.to_dict() == expected_rule

    @pytest.mark.parametrize(
        "ip_ver_addrs",
        [
            (
                "198.51.100.256/24",
                "198.51.100.2/33",
            ),
            (
                "2001:db8:a::ffff0/64",
                "2001:db8:a::/129",
            ),
            ("invalid_ip_string", "invalid_ip_string2"),
        ],
        ids=["ipv4", "ipv6", "invalid_format"],
    )
    def test_invalid_ip_address(self, ip_ver_addrs):
        ip_from, ip_to = ip_ver_addrs
        rule = _create_route_rule_dict(ip_from, ip_to, 50, 103)
        with pytest.raises(NmstateValueError):
            RouteRuleEntry(rule)

    def test_invalid_ipv6_address(self):
        rule = _create_route_rule_dict("256.0.0.0/24", "1.1.1.1/33", 50, 103)
        with pytest.raises(NmstateValueError):
            RouteRuleEntry(rule)

    def test_sort_route_rules(self):
        rules = [
            _create_route_rule("198.51.100.1/24", "192.0.2.1", 50, 103),
            _create_route_rule("198.51.100.0/24", "192.0.2.1", 50, 103),
            _create_route_rule("198.51.100.0/24", "192.0.2.1", 10, 103),
        ]
        expected_rules = [
            _create_route_rule("198.51.100.0/24", "192.0.2.1", 10, 103),
            _create_route_rule("198.51.100.0/24", "192.0.2.1", 50, 103),
            _create_route_rule("198.51.100.1/24", "192.0.2.1", 50, 103),
        ]
        assert expected_rules == sorted(rules)


class TestRouteRuleState:
    def _gen_ifaces(self):
        return gen_two_static_ip_ifaces(
            IPV4_ROUTE_IFACE_NAME, IPV6_ROUTE_IFACE_NAME
        )

    def _gen_route_state(self, ifaces):
        return RouteState(
            ifaces,
            {},
            {
                Route.CONFIG: [
                    gen_ipv4_route().to_dict(),
                    gen_ipv6_route().to_dict(),
                ]
            },
        )

    def test_verify_sort_rules(self):
        ifaces = self._gen_ifaces()
        state = RouteRuleState(
            self._gen_route_state(ifaces),
            {
                RouteRule.CONFIG: [
                    _gen_ipv4_route_rule().to_dict(),
                    _gen_ipv6_route_rule().to_dict(),
                ]
            },
            {},
        )
        state.verify(
            {
                RouteRule.CONFIG: [
                    _gen_ipv6_route_rule().to_dict(),
                    _gen_ipv4_route_rule().to_dict(),
                ]
            }
        )

    def test_gen_metatada(self):
        ifaces = self._gen_ifaces()
        route_state = self._gen_route_state(ifaces)
        route_rule_state = RouteRuleState(
            route_state,
            {
                RouteRule.CONFIG: [
                    _gen_ipv4_route_rule().to_dict(),
                    _gen_ipv6_route_rule().to_dict(),
                ]
            },
            {},
        )
        ifaces.gen_route_rule_metadata(route_rule_state, route_state)

        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.ROUTE_RULES_METADATA
        ] == [_gen_ipv4_route_rule().to_dict()]
        assert ipv6_iface.to_dict()[Interface.IPV6][
            BaseIface.ROUTE_RULES_METADATA
        ] == [_gen_ipv6_route_rule().to_dict()]

    def test_route_rule_to_unknow_route_table(self):
        ifaces = self._gen_ifaces()
        route_state = RouteState(
            ifaces,
            {},
            {Route.CONFIG: [gen_ipv4_route().to_dict()]},
        )
        route_rule_state = RouteRuleState(
            route_state,
            {
                RouteRule.CONFIG: [
                    _gen_ipv4_route_rule().to_dict(),
                    _gen_ipv6_route_rule().to_dict(),
                ]
            },
            {},
        )
        with pytest.raises(NmstateValueError):
            ifaces.gen_route_rule_metadata(route_rule_state, route_state)

    def test_discard_rule_to_unknown_table_when_merging(self):
        ifaces = self._gen_ifaces()
        route_state = RouteState(
            ifaces,
            {},
            {Route.CONFIG: [gen_ipv4_route().to_dict()]},
        )
        route_rule_state = RouteRuleState(
            route_state,
            {},
            {
                RouteRule.CONFIG: [
                    _gen_ipv4_route_rule().to_dict(),
                    _gen_ipv6_route_rule().to_dict(),
                ]
            },
        )
        ifaces.gen_route_rule_metadata(route_rule_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]
        ipv6_iface = ifaces.all_kernel_ifaces[IPV6_ROUTE_IFACE_NAME]

        assert ipv4_iface.to_dict()[Interface.IPV4][
            BaseIface.ROUTE_RULES_METADATA
        ] == [_gen_ipv4_route_rule().to_dict()]
        assert (
            BaseIface.ROUTE_RULES_METADATA
            not in ipv6_iface.to_dict()[Interface.IPV6]
        )

    def test_clear_rules_with_ip_from_using_wildcard(self):
        ifaces = self._gen_ifaces()
        route_state = RouteState(
            ifaces,
            {},
            {Route.CONFIG: [gen_ipv4_route().to_dict()]},
        )
        route_rule_state = RouteRuleState(
            route_state,
            {
                RouteRule.CONFIG: [
                    {
                        RouteRule.IP_FROM: IPV4_ROUTE_RULE_FROM,
                        RouteRule.STATE: RouteRule.STATE_ABSENT,
                    }
                ]
            },
            {
                RouteRule.CONFIG: [
                    {
                        RouteRule.IP_FROM: IPV4_ROUTE_RULE_FROM,
                        RouteRule.IP_TO: "192.0.2.1",
                        RouteRule.PRIORITY: 5000,
                    },
                    {
                        RouteRule.IP_FROM: IPV4_ROUTE_RULE_FROM,
                        RouteRule.IP_TO: IPV4_ROUTE_RULE_TO,
                        RouteRule.PRIORITY: 3000,
                    },
                ]
            },
        )
        ifaces.gen_route_rule_metadata(route_rule_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]

        assert (
            BaseIface.ROUTE_RULES_METADATA
            not in ipv4_iface.to_dict()[Interface.IPV4]
        )

    def test_clear_rules_with_ip_to_using_wildcard(self):
        ifaces = self._gen_ifaces()
        route_state = RouteState(
            ifaces,
            {},
            {Route.CONFIG: [gen_ipv4_route().to_dict()]},
        )
        route_rule_state = RouteRuleState(
            route_state,
            {
                RouteRule.CONFIG: [
                    {
                        RouteRule.IP_TO: IPV4_ROUTE_RULE_TO,
                        RouteRule.STATE: RouteRule.STATE_ABSENT,
                    }
                ]
            },
            {
                RouteRule.CONFIG: [
                    {
                        RouteRule.IP_FROM: IPV4_ROUTE_RULE_FROM,
                        RouteRule.IP_TO: IPV4_ROUTE_RULE_TO,
                        RouteRule.PRIORITY: 5000,
                    },
                    {
                        RouteRule.IP_FROM: "192.0.2.0/24",
                        RouteRule.IP_TO: IPV4_ROUTE_RULE_TO,
                        RouteRule.PRIORITY: 3000,
                    },
                ]
            },
        )
        ifaces.gen_route_rule_metadata(route_rule_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]

        assert (
            BaseIface.ROUTE_RULES_METADATA
            not in ipv4_iface.to_dict()[Interface.IPV4]
        )

    def test_clear_rules_with_priority_using_wildcard(self):
        ifaces = self._gen_ifaces()
        route_state = RouteState(
            ifaces,
            {},
            {Route.CONFIG: [gen_ipv4_route().to_dict()]},
        )
        route_rule_state = RouteRuleState(
            route_state,
            {
                RouteRule.CONFIG: [
                    {
                        RouteRule.PRIORITY: 5000,
                        RouteRule.STATE: RouteRule.STATE_ABSENT,
                    }
                ]
            },
            {
                RouteRule.CONFIG: [
                    {
                        RouteRule.IP_TO: IPV4_ROUTE_RULE_TO,
                        RouteRule.PRIORITY: 5000,
                    },
                    {
                        RouteRule.IP_TO: IPV4_ROUTE_RULE_TO,
                        RouteRule.PRIORITY: 5000,
                    },
                ]
            },
        )
        ifaces.gen_route_rule_metadata(route_rule_state, route_state)
        ipv4_iface = ifaces.all_kernel_ifaces[IPV4_ROUTE_IFACE_NAME]

        assert (
            BaseIface.ROUTE_RULES_METADATA
            not in ipv4_iface.to_dict()[Interface.IPV4]
        )


def _create_route_rule(ip_from, ip_to, priority, table):
    return RouteRuleEntry(
        _create_route_rule_dict(ip_from, ip_to, priority, table)
    )


def _create_route_rule_dict(ip_from, ip_to, priority, table):
    return {
        RouteRule.IP_FROM: ip_from,
        RouteRule.IP_TO: ip_to,
        RouteRule.PRIORITY: priority,
        RouteRule.ROUTE_TABLE: table,
    }


def _gen_ipv4_route_rule():
    return _create_route_rule(
        IPV4_ROUTE_RULE_FROM, IPV4_ROUTE_RULE_TO, 5000, IPV4_ROUTE_TABLE_ID
    )


def _gen_ipv6_route_rule():
    return _create_route_rule(
        IPV6_ROUTE_RULE_FROM, IPV6_ROUTE_RULE_TO, 5001, IPV6_ROUTE_TABLE_ID
    )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

import pytest

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6

from libnmstate.ifaces.base_iface import IPState

from ..testlib.constants import IPV4_ADDRESSES
from ..testlib.constants import IPV6_ADDRESS1
from ..testlib.constants import IPV6_ADDRESS1_FULL
from ..testlib.constants import IPV6_LINK_LOCAL_ADDRESS1
from ..testlib.constants import IPV6_ADDRESSES


parametrize_ip_ver = pytest.mark.parametrize(
    "ip_ver",
    [
        (
            Interface.IPV4,
            {
                InterfaceIPv4.ENABLED: True,
                InterfaceIPv4.ADDRESS: deepcopy(IPV4_ADDRESSES),
            },
        ),
        (
            Interface.IPV6,
            {
                InterfaceIPv6.ENABLED: True,
                InterfaceIPv6.ADDRESS: deepcopy(IPV6_ADDRESSES),
            },
        ),
    ],
    ids=["ipv4", "ipv6"],
)

parametrize_ip_ver_dynamic = pytest.mark.parametrize(
    "ip_ver_dynamic",
    [
        (
            Interface.IPV4,
            {
                InterfaceIPv4.ENABLED: True,
                InterfaceIPv4.DHCP: True,
                InterfaceIPv4.ADDRESS: [],
                InterfaceIPv4.AUTO_DNS: True,
                InterfaceIPv4.AUTO_GATEWAY: True,
                InterfaceIPv4.AUTO_ROUTES: True,
            },
        ),
        (
            Interface.IPV6,
            {
                InterfaceIPv6.ENABLED: True,
                InterfaceIPv6.DHCP: False,
                InterfaceIPv6.ADDRESS: [],
                InterfaceIPv6.AUTOCONF: True,
                InterfaceIPv6.AUTO_DNS: True,
                InterfaceIPv6.AUTO_GATEWAY: True,
                InterfaceIPv6.AUTO_ROUTES: True,
            },
        ),
        (
            Interface.IPV6,
            {
                InterfaceIPv6.ENABLED: True,
                InterfaceIPv6.DHCP: True,
                InterfaceIPv6.ADDRESS: [],
                InterfaceIPv6.AUTOCONF: False,
                InterfaceIPv6.AUTO_DNS: True,
                InterfaceIPv6.AUTO_GATEWAY: True,
                InterfaceIPv6.AUTO_ROUTES: True,
            },
        ),
        (
            Interface.IPV6,
            {
                InterfaceIPv6.ENABLED: True,
                InterfaceIPv6.DHCP: True,
                InterfaceIPv6.ADDRESS: [],
                InterfaceIPv6.AUTOCONF: True,
                InterfaceIPv6.AUTO_DNS: True,
                InterfaceIPv6.AUTO_GATEWAY: True,
                InterfaceIPv6.AUTO_ROUTES: True,
            },
        ),
    ],
    ids=["dhcpv4", "ipv6-ra", "dhcpv6", "dhcpv6+ra"],
)


class TestIPState:
    def test_discard_other_option_when_disabled(self):
        ip_state = IPState(
            Interface.IPV6, {InterfaceIPv6.ENABLED: False, "foo_a": "foo_b"}
        )
        assert ip_state.to_dict() == {
            InterfaceIP.ENABLED: False,
        }

    @parametrize_ip_ver
    def test_sort_address(self, ip_ver):
        family, ip_info = ip_ver
        ip_state1 = IPState(family, ip_info)
        ip_info2 = deepcopy(ip_info)
        ip_info2[InterfaceIP.ADDRESS].reverse()
        ip_state2 = IPState(family, ip_info2)
        assert ip_state1.to_dict() == ip_state2.to_dict()

    def test_ipv6_non_abbreviated_address(self):
        ip_state = IPState(
            Interface.IPV6,
            {
                InterfaceIPv6.ENABLED: True,
                InterfaceIPv6.ADDRESS: [
                    {
                        InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1_FULL,
                        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                    }
                ],
            },
        )

        assert ip_state.to_dict() == {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.ADDRESS: [
                {
                    InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                    InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                }
            ],
        }

    @parametrize_ip_ver_dynamic
    def test_default_dynamic_options_when_undefined(self, ip_ver_dynamic):
        ip_info = {
            InterfaceIP.ENABLED: True,
        }
        family, dynamic_options = ip_ver_dynamic
        ip_info.update(dynamic_options)
        ip_info.pop(InterfaceIP.AUTO_DNS)
        ip_info.pop(InterfaceIP.AUTO_ROUTES)
        ip_info.pop(InterfaceIP.AUTO_GATEWAY)

        expected_ip_info = deepcopy(ip_info)
        expected_ip_info.update(dynamic_options)

        ip_state = IPState(family, ip_info)

        assert ip_state.to_dict() == expected_ip_info

    @parametrize_ip_ver_dynamic
    def test_discard_address_when_dynamic(self, ip_ver_dynamic):
        ip_info = {}
        family, dynamic_options = ip_ver_dynamic
        ip_info.update(dynamic_options)
        expected_ip_info = deepcopy(ip_info)
        ip_info[InterfaceIP.ADDRESS] = (
            IPV4_ADDRESSES if family == InterfaceIPv4 else IPV6_ADDRESSES
        )

        ip_state = IPState(family, ip_info)

        assert ip_state.to_dict() == expected_ip_info

    @parametrize_ip_ver
    def test_remove_link_local_address(self, ip_ver):
        family, ip_info = ip_ver
        expected_ip_info = deepcopy(ip_info)
        if family == Interface.IPV6:
            ip_info[InterfaceIPv6.ADDRESS].append(
                {
                    InterfaceIPv6.ADDRESS_IP: IPV6_LINK_LOCAL_ADDRESS1,
                    InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                }
            )

        ip_state = IPState(Interface.IPV6, ip_info)
        ip_state.remove_link_local_address()

        assert ip_state.to_dict() == expected_ip_info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import datetime
import logging

from libnmstate.error import NmstateInternalError
from libnmstate.error import NmstateTimeoutError

from .common import NM
from .common import GLib
from .common import Gio

# Interval for idle checker to check on whether timeout should trigger since
# last finish async action.
IDLE_CHECK_INTERNAL = 5

# libnm dbus connection has reply timeout 25 seconds.
IDLE_TIMEOUT = 25

# NetworkManage is using dbus in libnm while the dbus has limitation on
# maximum number of pending replies per connection.(RHEL/CentOS 8 is 1024)
# Hence limit the synchronous queue size
SLOW_ASYNC_QUEUE_SIZE = 100
FAST_ASYNC_QUEUE_SIZE = 300


class NmContext:
    def __init__(self):
        self._client = NM.Client.new(cancellable=None)
        self._context = self._client.get_main_context()
        self._quitting = False
        self._cancellable = None
        self._error = None
        self._timeout_source = None
        self._last_async_finish_time = None
        self._fast_queue = None
        self._slow_queue = None
        self._init_queue()
        self._init_cancellable()

    def _init_client(self):
        self._client = NM.Client.new(cancellable=None)
        self._context = self._client.get_main_context()

    def _init_queue(self):
        self._fast_queue = set()
        self._slow_queue = set()

    def _init_cancellable(self):
        self._cancellable = Gio.Cancellable.new()

    @property
    def cancellable(self):
        return self._cancellable

    @property
    def client(self):
        if self._quitting:
            return None
        return self._client

    @property
    def context(self):
        if not self._context:
            raise NmstateInternalError(
                "BUG: Accessing MainContext while it is None"
            )
        return self._context

    def refresh_content(self):
        if self.context:
            while self.context.iteration(False):
                pass

    def clean_up(self):
        if self._cancellable:
            self._cancellable.cancel()
        self._del_timeout()
        self._del_client()
        self._context = None
        self._cancellable = None

    def _del_client(self):
        if self._client:
            is_done = []
            is_timeout = []
            self._client.get_context_busy_watcher().weak_ref(
                lambda: is_done.append(1)
            )
            self._client = None
            self._quitting = True

            self.refresh_content()

            if not is_done:
                timeout_source = GLib.timeout_source_new(50)
                try:
                    timeout_source.set_callback(lambda x: is_timeout.append(1))
                    timeout_source.attach(self.context)
                    while not is_done and not is_timeout:
                        self.context.iteration(True)
                finally:
                    timeout_source.destroy()
            if not is_done:
                logging.error("BUG: NM.Client is not cleaned")
            self._context = None

    def _del_timeout(self):
        if self._timeout_source:
            self._timeout_source.destroy()
            self._timeout_source = None

    def register_async(self, action, fast=False):
        """
        Register action(string) to wait list.
        Set fast as True if requested action does not require too much time,
        for example: profile modification.
        """
        queue = self._fast_queue if fast else self._slow_queue
        max_queue = FAST_ASYNC_QUEUE_SIZE if fast else SLOW_ASYNC_QUEUE_SIZE
        if len(queue) >= max_queue:
            logging.debug(
                f"Async queue({max_queue}) full, waiting all existing actions "
                "to be finished before registering more async action"
            )
            # TODO: No need to wait all finish, should continue when the queue
            #       is considerably empty and ready for new async action.
            self.wait_all_finish()

        if action in self._fast_queue or action in self._slow_queue:
            raise NmstateInternalError(
                f"BUG: An existing actions {action} is already registered"
            )

        logging.debug(f"Async action: {action} started")
        queue.add(action)

    def finish_async(self, action, suppress_log=False):
        """
        Mark action(string) as finished.
        """
        self._last_async_finish_time = datetime.datetime.now()
        if not suppress_log:
            logging.debug(f"Async action: {action} finished")
        self._fast_queue.discard(action)
        self._slow_queue.discard(action)

    def _action_all_finished(self):
        return not (len(self._fast_queue) or len(self._slow_queue))

    def _idle_timeout_cb(self, _user_data):
        if self._error or self._action_all_finished():
            return GLib.SOURCE_REMOVE
        idle_time = datetime.datetime.now() - self._last_async_finish_time
        if idle_time > datetime.timedelta(seconds=IDLE_TIMEOUT):
            remaining_actions = self._slow_queue | self._fast_queue
            self.fail(
                NmstateTimeoutError(f"Action {remaining_actions} timeout")
            )
            return GLib.SOURCE_REMOVE
        else:
            return GLib.SOURCE_CONTINUE

    def is_cancelled(self):
        return self._cancellable.is_cancelled()

    def fail(self, exception):
        if not self._cancellable.is_cancelled():
            if self._error:
                logging.error(
                    f"BUG: There is already a exception assigned: "
                    f"existing: {self._error}, new exception {exception}"
                )
            self.cancellable.cancel()
            self._del_timeout()
            self._error = exception

    def wait_all_finish(self):
        """
        Block till all async actions been marked as finished via
        `finish_async()` or anyone failed by `fail()`.
        """
        self._last_async_finish_time = datetime.datetime.now()
        if not self._action_all_finished():
            self._timeout_source = GLib.timeout_source_new(
                IDLE_CHECK_INTERNAL * 1000
            )
            user_data = None
            self._timeout_source.set_callback(self._idle_timeout_cb, user_data)
            self._timeout_source.attach(self._context)

            while not self._action_all_finished() and not self._error:
                self.context.iteration(True)
            self._del_timeout()

        if self._error:
            # The queue and error should be flush and perpare for another run
            self._cancellable.cancel()
            self.refresh_content()
            self._init_queue()
            self._init_cancellable()
            tmp_error = self._error
            self._error = None
            # pylint: disable=raising-bad-type
            raise tmp_error
            # pylint: enable=raising-bad-type



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import Bond
from libnmstate.schema import InterfaceType

from .base_iface import NisporPluginBaseIface

SUPPORTED_BOND_OPTIONS = (
    "ad_actor_sys_prio",
    "ad_actor_system",
    "ad_select",
    "ad_user_port_key",
    "all_subordinates_active",
    "arp_all_targets",
    "arp_interval",
    "arp_ip_target",
    "arp_validate",
    "downdelay",
    "fail_over_mac",
    "lacp_rate",
    "lp_interval",
    "miimon",
    "min_links",
    "num_grat_arp",
    "num_unsol_na",
    "packets_per_subordinate",
    "primary",
    "primary_reselect",
    "resend_igmp",
    "tlb_dynamic_lb",
    "updelay",
    "use_carrier",
    "xmit_hash_policy",
)


class NisporPluginBondIface(NisporPluginBaseIface):
    @property
    def type(self):
        return InterfaceType.BOND

    def to_dict(self, config_only):
        info = super().to_dict(config_only)
        info[Bond.CONFIG_SUBTREE] = {
            Bond.MODE: self._np_iface.mode,
            Bond.PORT: self._np_iface.subordinates,
            Bond.OPTIONS_SUBTREE: {},
        }
        bond_opts = info[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE]
        for opt_name in SUPPORTED_BOND_OPTIONS:
            value = getattr(self._np_iface, opt_name)
            if opt_name == "all_subordinates_active":
                # The sysfs is using `all_slave_active` name
                opt_name = "all_slaves_active"
            elif opt_name == "packets_per_subordinate":
                opt_name = "packets_per_slave"
            if value is not None:
                bond_opts[opt_name] = value
        return info



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType

from ..testlib import statelib
from ..testlib import cmdlib


BRIDGE0 = "brtest0"


@pytest.fixture
def ovs_unmanaged_bridge():
    cmdlib.exec_cmd(f"ovs-vsctl add-br {BRIDGE0}".split())
    yield
    cmdlib.exec_cmd(f"ovs-vsctl del-br {BRIDGE0}".split())


@pytest.mark.tier1
def test_do_not_show_unmanaged_ovs_bridge(ovs_unmanaged_bridge):
    # The output should only contains the OVS internal interface
    ovs_internal_iface = statelib.show_only((BRIDGE0,))[Interface.KEY][0]
    assert ovs_internal_iface[Interface.TYPE] == InterfaceType.OVS_INTERFACE



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import os


from .testlib import cmdlib


RC_SUCCESS = 0


def test_edit_abort():
    runenv = dict(os.environ)
    env = {"EDITOR": "false"}

    runenv.update(env)

    cmds = ["nmstatectl", "edit", "lo"]
    ret = cmdlib.exec_cmd(cmds, env=runenv)
    rc, out, err = ret

    assert_rc(rc, os.EX_DATAERR, ret)


def test_edit_no_change_eth1():
    runenv = dict(os.environ)
    env = {"EDITOR": "touch"}

    runenv.update(env)

    cmds = ["nmstatectl", "edit", "eth1"]
    ret = cmdlib.exec_cmd(cmds, env=runenv)
    rc, out, err = ret

    assert_rc(rc, RC_SUCCESS, ret)


def assert_rc(actual, expected, return_tuple):
    assert actual == expected, "rc={}, out={}, err={}".format(*return_tuple)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from unittest import mock

from libnmstate import nm
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType


@pytest.fixture(scope="module")
def NM_mock():
    saved_api2nm_map = nm.translator.Api2Nm._iface_types_map
    saved_nm2api_map = nm.translator.Nm2Api._iface_types_map
    nm.translator.Api2Nm._iface_types_map = None
    nm.translator.Nm2Api._iface_types_map = None

    with mock.patch.object(nm.translator, "NM") as m:
        yield m

    nm.translator.Api2Nm._iface_types_map = saved_api2nm_map
    nm.translator.Nm2Api._iface_types_map = saved_nm2api_map


def test_api2nm_iface_type_map(NM_mock):
    ovs_interface_setting = NM_mock.SETTING_OVS_INTERFACE_SETTING_NAME
    map = nm.translator.Api2Nm.get_iface_type_map()

    expected_map = {
        InterfaceType.ETHERNET: NM_mock.SETTING_WIRED_SETTING_NAME,
        InterfaceType.BOND: NM_mock.SETTING_BOND_SETTING_NAME,
        InterfaceType.DUMMY: NM_mock.SETTING_DUMMY_SETTING_NAME,
        InterfaceType.OVS_BRIDGE: NM_mock.SETTING_OVS_BRIDGE_SETTING_NAME,
        InterfaceType.OVS_PORT: NM_mock.SETTING_OVS_PORT_SETTING_NAME,
        InterfaceType.OVS_INTERFACE: ovs_interface_setting,
        InterfaceType.TEAM: NM_mock.SETTING_TEAM_SETTING_NAME,
        InterfaceType.VLAN: NM_mock.SETTING_VLAN_SETTING_NAME,
        InterfaceType.LINUX_BRIDGE: NM_mock.SETTING_BRIDGE_SETTING_NAME,
        InterfaceType.VXLAN: NM_mock.SETTING_VXLAN_SETTING_NAME,
        InterfaceType.VRF: NM_mock.SETTING_VRF_SETTING_NAME,
        InterfaceType.INFINIBAND: NM_mock.SETTING_INFINIBAND_SETTING_NAME,
        InterfaceType.MAC_VLAN: NM_mock.SETTING_MACVLAN_SETTING_NAME,
        InterfaceType.MAC_VTAP: NM_mock.SETTING_MACVLAN_SETTING_NAME,
    }

    assert map == expected_map


def test_api2nm_get_iface_type(NM_mock):
    nm_type = nm.translator.Api2Nm.get_iface_type(InterfaceType.ETHERNET)
    assert NM_mock.SETTING_WIRED_SETTING_NAME == nm_type


@mock.patch.object(
    nm.translator.Api2Nm, "get_iface_type", staticmethod(lambda t: t)
)
def test_api2nm_bond_options():
    bond_options = {
        Interface.NAME: "bond99",
        Interface.TYPE: InterfaceType.BOND,
        Interface.STATE: InterfaceState.UP,
        Bond.CONFIG_SUBTREE: {
            Bond.MODE: BondMode.ROUND_ROBIN,
            Bond.OPTIONS_SUBTREE: {"miimon": 120},
        },
    }
    nm_bond_options = nm.translator.Api2Nm.get_bond_options(bond_options)

    assert {"miimon": 120, Bond.MODE: BondMode.ROUND_ROBIN} == nm_bond_options


def test_nm2api_common_device_info(NM_mock):
    NM_mock.DeviceState.ACTIVATED = 100
    NM_mock.DeviceState.IP_CONFIG = 70
    nm.common.NM.DeviceState.DISCONNECTED = 30
    devinfo = {
        "name": "devname",
        "type_id": "devtypeid",
        "type_name": "devtypename",
        "state": nm.common.NM.DeviceState.DISCONNECTED,
    }
    info = nm.translator.Nm2Api.get_common_device_info(devinfo)

    expected_info = {
        Interface.NAME: "devname",
        Interface.STATE: InterfaceState.DOWN,
        Interface.TYPE: InterfaceType.UNKNOWN,
    }
    assert expected_info == info


def test_nm2api_bond_info():
    port_mock = [mock.MagicMock(), mock.MagicMock()]
    bondinfo = {
        Bond.PORT: port_mock,
        Bond.OPTIONS_SUBTREE: {Bond.MODE: BondMode.ROUND_ROBIN, "miimon": 120},
    }
    info = nm.translator.Nm2Api.get_bond_info(bondinfo)

    expected_info = {
        Bond.CONFIG_SUBTREE: {
            Bond.MODE: BondMode.ROUND_ROBIN,
            Bond.PORT: [
                port_mock[0].props.interface,
                port_mock[1].props.interface,
            ],
            Bond.OPTIONS_SUBTREE: {"miimon": 120},
        }
    }
    assert expected_info == info


def test_iface_admin_state(NM_mock):
    NM_mock.DeviceState.ACTIVATED = 100
    NM_mock.DeviceState.IP_CONFIG = 70
    NM_mock.DeviceState.IP_CHECK = 80
    admin_state = nm.translator.Nm2Api.get_iface_admin_state(
        NM_mock.DeviceState.IP_CHECK
    )

    assert InterfaceState.UP == admin_state



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.error import NmstateVerificationError
from libnmstate.schema import Ethernet
from libnmstate.schema import LinuxBridge
from libnmstate.schema import VLAN
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType

from libnmstate.ifaces.ifaces import Ifaces
from libnmstate.ifaces.ifaces import OvsBridgeIface
from libnmstate.state import state_match

from .testlib.bridgelib import LINUX_BRIDGE_IFACE_NAME
from .testlib.bridgelib import PORT1_IFACE_NAME
from .testlib.bridgelib import PORT2_IFACE_NAME
from .testlib.bridgelib import gen_bridge_iface_info
from .testlib.constants import MAC_ADDRESS1
from .testlib.ifacelib import gen_foo_iface_info_static_ip
from .testlib.ifacelib import gen_foo_iface_info
from .testlib.ovslib import OVS_BRIDGE_IFACE_NAME
from .testlib.ovslib import OVS_IFACE_NAME
from .testlib.ovslib import gen_ovs_bridge_info

FOO1_IFACE_NAME = "foo1"
FOO2_IFACE_NAME = "foo2"
FOO3_IFACE_NAME = "foo3"
FOO4_IFACE_NAME = "foo4"
PARENT_IFACE_NAME = "parent"
CHILD_IFACE_NAME = "child"


class TestIfaces:
    def _gen_iface_infos(self):
        iface1_info = gen_foo_iface_info_static_ip()
        iface1_info[Interface.NAME] = FOO1_IFACE_NAME
        iface2_info = gen_foo_iface_info_static_ip()
        iface2_info[Interface.NAME] = FOO2_IFACE_NAME
        return [iface1_info, iface2_info]

    def test_duplicate_interface_name(self):
        iface_infos = self._gen_iface_infos()
        iface_infos.append(deepcopy(iface_infos[0]))
        with pytest.raises(NmstateValueError):
            Ifaces(des_iface_infos=iface_infos, cur_iface_infos=[])

    def test_init_cur_iface_infos_only(self):
        ifaces = Ifaces(
            des_iface_infos=[], cur_iface_infos=self._gen_iface_infos()
        )

        for iface_info in self._gen_iface_infos():
            cur_iface = ifaces.get_cur_iface(
                iface_info[Interface.NAME], iface_info[Interface.TYPE]
            )
            assert cur_iface.original_dict == iface_info

    def test_init_des_iface_infos_only(self):
        ifaces = Ifaces(
            cur_iface_infos=[], des_iface_infos=self._gen_iface_infos()
        )

        assert [
            i.original_dict for i in ifaces.all_ifaces()
        ] == self._gen_iface_infos()

    def test_add_new_iface(self):
        cur_iface_infos = self._gen_iface_infos()
        new_iface_info = gen_foo_iface_info()
        new_iface_info[Interface.NAME] = FOO3_IFACE_NAME

        ifaces = Ifaces([new_iface_info], cur_iface_infos)
        new_iface = ifaces.all_kernel_ifaces[FOO3_IFACE_NAME]
        iface1 = ifaces.all_kernel_ifaces[FOO1_IFACE_NAME]
        iface2 = ifaces.all_kernel_ifaces[FOO2_IFACE_NAME]

        assert (
            len(list(ifaces.all_kernel_ifaces.keys()))
            == len(self._gen_iface_infos()) + 1
        )
        assert new_iface.is_desired
        assert not iface1.is_desired
        assert not iface2.is_desired

    def test_edit_existing_iface(self):
        cur_iface_infos = self._gen_iface_infos()
        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0][Interface.MAC] = MAC_ADDRESS1
        expected_iface_info = deepcopy(des_iface_infos[0])

        ifaces = Ifaces(des_iface_infos, cur_iface_infos)
        edit_iface = ifaces.all_kernel_ifaces[
            des_iface_infos[0][Interface.NAME]
        ]

        assert state_match(expected_iface_info, edit_iface.to_dict())
        assert edit_iface.is_desired

    def test_mark_iface_as_absent(self):
        cur_iface_infos = self._gen_iface_infos()
        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0][Interface.STATE] = InterfaceState.ABSENT
        expected_iface_info = deepcopy(des_iface_infos[0])

        ifaces = Ifaces(des_iface_infos, cur_iface_infos)
        edit_iface = ifaces.all_kernel_ifaces[
            des_iface_infos[0][Interface.NAME]
        ]

        assert state_match(expected_iface_info, edit_iface.to_dict())
        assert edit_iface.is_desired
        assert edit_iface.is_absent

    def test_validate_unknown_port(self):
        cur_iface_infos = self._gen_iface_infos()
        des_iface_info = gen_bridge_iface_info()

        with pytest.raises(NmstateValueError):
            Ifaces([des_iface_info], cur_iface_infos)

    def test_validate_overbooked_port(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        des_iface_info1 = gen_bridge_iface_info()
        des_iface_info2 = gen_bridge_iface_info()
        des_iface_info2[Interface.NAME] = "another_bridge"

        with pytest.raises(NmstateValueError):
            Ifaces([des_iface_info1, des_iface_info2], cur_iface_infos)

    def test_remove_unknown_interfaces(self):
        des_iface_infos = self._gen_iface_infos()
        cur_iface_info = {
            Interface.NAME: FOO3_IFACE_NAME,
            Interface.TYPE: InterfaceType.UNKNOWN,
            Interface.STATE: InterfaceState.UP,
        }

        ifaces = Ifaces(des_iface_infos, [cur_iface_info])
        assert cur_iface_info not in ifaces.all_ifaces()

    def test_mark_port_as_changed_if_controller_marked_as_absent(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        cur_iface_infos.append(gen_bridge_iface_info())
        des_iface_info = gen_bridge_iface_info()
        des_iface_info[Interface.STATE] = InterfaceState.ABSENT

        ifaces = Ifaces([des_iface_info], cur_iface_infos)

        port_iface1 = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port_iface2 = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]
        controller_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]

        assert port_iface1.is_changed
        assert port_iface2.is_changed
        assert port_iface1.is_up
        assert port_iface2.is_up
        assert port_iface1.controller is None
        assert port_iface2.controller is None
        assert controller_iface.is_desired
        assert controller_iface.is_absent

    def test_mark_port_as_changed_when_controller_changed_port_list(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        cur_iface_infos.append(gen_bridge_iface_info())
        des_iface_info = gen_bridge_iface_info()
        des_iface_info[LinuxBridge.CONFIG_SUBTREE][
            LinuxBridge.PORT_SUBTREE
        ].pop()

        ifaces = Ifaces([des_iface_info], cur_iface_infos)

        port_iface1 = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port_iface2 = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]
        controller_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]

        assert not port_iface1.is_changed
        assert port_iface2.is_changed
        assert port_iface1.is_up
        assert port_iface2.is_up
        assert port_iface1.controller == LINUX_BRIDGE_IFACE_NAME
        assert port_iface2.controller is None
        assert controller_iface.is_desired

    def test_mark_port_as_changed_when_a_port_to_new_controller(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        des_iface_info = gen_bridge_iface_info()

        ifaces = Ifaces([des_iface_info], cur_iface_infos)

        port_iface1 = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port_iface2 = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]
        controller_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]

        assert port_iface1.is_changed
        assert port_iface2.is_changed
        assert port_iface1.is_up
        assert port_iface2.is_up
        assert port_iface1.controller == LINUX_BRIDGE_IFACE_NAME
        assert port_iface2.controller == LINUX_BRIDGE_IFACE_NAME
        assert controller_iface.is_desired

    def test_mark_port_as_changed_when_controller_changed_port_config(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        cur_iface_infos.append(gen_bridge_iface_info())
        des_iface_info = gen_bridge_iface_info()
        des_iface_info[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.PORT_SUBTREE][
            0
        ][LinuxBridge.Port.STP_PATH_COST] += 1

        ifaces = Ifaces([des_iface_info], cur_iface_infos)

        port_iface1 = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port_iface2 = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]
        controller_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]

        assert port_iface1.is_changed
        assert not port_iface2.is_changed
        assert port_iface1.is_up
        assert port_iface2.is_up
        assert port_iface1.controller == LINUX_BRIDGE_IFACE_NAME
        assert port_iface2.controller == LINUX_BRIDGE_IFACE_NAME
        assert controller_iface.is_desired

    def test_mark_child_as_absent_when_parent_is_marked_as_absent(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PARENT_IFACE_NAME
        child_iface_info = gen_foo_iface_info(iface_type=InterfaceType.VLAN)
        child_iface_info[Interface.NAME] = CHILD_IFACE_NAME
        child_iface_info[VLAN.CONFIG_SUBTREE] = {
            VLAN.ID: 101,
            VLAN.BASE_IFACE: PARENT_IFACE_NAME,
        }
        cur_iface_infos.append(child_iface_info)

        des_iface_info = gen_foo_iface_info()
        des_iface_info[Interface.NAME] = PARENT_IFACE_NAME
        des_iface_info[Interface.STATE] = InterfaceState.ABSENT

        ifaces = Ifaces([des_iface_info], cur_iface_infos)

        child_iface = ifaces.all_kernel_ifaces[CHILD_IFACE_NAME]
        parent_iface = ifaces.all_kernel_ifaces[PARENT_IFACE_NAME]
        other_iface = ifaces.all_kernel_ifaces[FOO2_IFACE_NAME]

        assert parent_iface.is_desired
        assert child_iface.is_changed
        assert not other_iface.is_changed
        assert parent_iface.is_absent
        assert child_iface.is_absent
        assert other_iface.is_up

    def test_mark_orphen_as_absent(self):
        """
        When OVS internal interface been removed from OVS bridge port list,
        the orphen interface should be marked as absent.
        """
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        cur_iface_infos.append(gen_ovs_bridge_info())
        child_iface_info = gen_foo_iface_info(
            iface_type=InterfaceType.OVS_INTERFACE
        )
        child_iface_info[Interface.NAME] = OVS_IFACE_NAME
        cur_iface_infos.append(child_iface_info)

        des_ovs_bridge_info = gen_ovs_bridge_info()
        ovs_bridge_iface = OvsBridgeIface(des_ovs_bridge_info, True)
        ovs_bridge_iface.remove_port(OVS_IFACE_NAME)
        des_iface_info = ovs_bridge_iface.to_dict()

        ifaces = Ifaces([des_iface_info], cur_iface_infos)

        ovs_iface = ifaces.all_kernel_ifaces[OVS_IFACE_NAME]
        bridge_iface = ifaces.get_iface(
            OVS_BRIDGE_IFACE_NAME, InterfaceType.OVS_BRIDGE
        )
        port1_iface = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port2_iface = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]

        assert bridge_iface.is_desired
        assert ovs_iface.is_changed
        assert not port1_iface.is_changed
        assert not port2_iface.is_changed
        assert bridge_iface.is_up
        assert port1_iface.is_up
        assert port2_iface.is_up
        assert ovs_iface.is_absent

    def test_validate_unknown_parent(self):
        child_iface_info = gen_foo_iface_info(iface_type=InterfaceType.VLAN)
        child_iface_info[VLAN.CONFIG_SUBTREE] = {
            VLAN.ID: 101,
            VLAN.BASE_IFACE: PARENT_IFACE_NAME,
        }
        with pytest.raises(NmstateValueError):
            Ifaces([child_iface_info], [])

    def test_state_to_edit_with_empty_desire(self):
        cur_iface_infos = self._gen_iface_infos()
        ifaces = Ifaces([], cur_iface_infos)

        assert ifaces.state_to_edit == []

    def test_state_to_edit_with_desire(self):
        cur_iface_infos = self._gen_iface_infos()
        des_iface_infos = self._gen_iface_infos()
        ifaces = Ifaces(des_iface_infos, cur_iface_infos)

        assert sorted(
            [i[Interface.NAME] for i in ifaces.state_to_edit]
        ) == sorted([FOO1_IFACE_NAME, FOO2_IFACE_NAME])

    def test_state_to_edit_with_changed_only(self):
        cur_iface_infos = self._gen_iface_infos()
        ifaces = Ifaces([], cur_iface_infos)

        for iface in ifaces.all_ifaces():
            iface.mark_as_changed()

        assert sorted(
            [i[Interface.NAME] for i in ifaces.state_to_edit]
        ) == sorted([FOO1_IFACE_NAME, FOO2_IFACE_NAME])

    def test_state_to_edit_with_ignored_iface(self):
        cur_iface_infos = self._gen_iface_infos()
        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0][Interface.STATE] = InterfaceState.IGNORE
        ifaces = Ifaces(des_iface_infos, cur_iface_infos)

        assert sorted([i[Interface.NAME] for i in ifaces.state_to_edit]) == [
            FOO2_IFACE_NAME
        ]

    def test_state_to_edit_with_ignored_subordinate(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        cur_iface_infos.append(gen_bridge_iface_info())
        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        des_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        des_iface_infos[1][Interface.STATE] = InterfaceState.IGNORE
        des_bridge_info = gen_bridge_iface_info()
        des_bridge_info[LinuxBridge.CONFIG_SUBTREE][
            LinuxBridge.PORT_SUBTREE
        ] = []
        des_iface_infos.append(des_bridge_info)

        ifaces = Ifaces(des_iface_infos, cur_iface_infos)

        assert sorted(
            [i[Interface.NAME] for i in ifaces.state_to_edit]
        ) == sorted([LINUX_BRIDGE_IFACE_NAME, PORT1_IFACE_NAME])

    def test_verify_desire_iface_not_found_in_current(self):
        cur_iface_infos = self._gen_iface_infos()
        cur_iface_infos.pop()

        des_iface_infos = self._gen_iface_infos()
        des_ifaces = Ifaces(des_iface_infos, [])

        with pytest.raises(NmstateVerificationError):
            des_ifaces.verify(cur_iface_infos)

    def test_verify_desire_iface_matches(self):
        cur_iface_infos = self._gen_iface_infos()

        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0].pop(Interface.TYPE)
        des_ifaces = Ifaces(des_iface_infos, cur_iface_infos)

        des_ifaces.verify(cur_iface_infos)

    def test_verify_desire_iface_not_match(self):
        cur_iface_infos = self._gen_iface_infos()

        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0][Interface.MAC] = MAC_ADDRESS1
        des_ifaces = Ifaces(des_iface_infos, cur_iface_infos)

        with pytest.raises(NmstateVerificationError):
            des_ifaces.verify(cur_iface_infos)

    def test_verify_with_ignored_interface(self):
        cur_iface_infos = self._gen_iface_infos()
        des_iface_infos = self._gen_iface_infos()
        des_iface_infos[0][Interface.STATE] = InterfaceState.IGNORE
        des_iface_infos[0]["foo"] = "bar"
        ifaces = Ifaces(des_iface_infos, cur_iface_infos)

        ifaces.verify(cur_iface_infos)

    def test_verify_with_ignored_subordinate(self):
        cur_iface_infos = [gen_foo_iface_info(), gen_foo_iface_info()]
        cur_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        cur_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        cur_iface_infos.append(gen_bridge_iface_info())
        des_iface_infos = [gen_foo_iface_info(), gen_foo_iface_info()]
        des_iface_infos[0][Interface.NAME] = PORT1_IFACE_NAME
        des_iface_infos[1][Interface.NAME] = PORT2_IFACE_NAME
        des_iface_infos[1][Interface.STATE] = InterfaceState.IGNORE
        des_bridge_info = gen_bridge_iface_info()
        des_bridge_info[LinuxBridge.CONFIG_SUBTREE][
            LinuxBridge.PORT_SUBTREE
        ].pop()
        des_iface_infos.append(des_bridge_info)

        ifaces = Ifaces(des_iface_infos, cur_iface_infos)
        ifaces.verify(cur_iface_infos)


class TestIfacesSriov:
    def test_vf_been_auto_include_as_desired(self):
        des_iface_infos = [gen_foo_iface_info(InterfaceType.ETHERNET)]
        des_iface_infos[0][Interface.NAME] = FOO1_IFACE_NAME
        des_iface_infos[0][Ethernet.CONFIG_SUBTREE] = {
            Ethernet.SRIOV_SUBTREE: {Ethernet.SRIOV.TOTAL_VFS: 2}
        }
        ifaces = Ifaces(des_iface_infos, [])
        assert len(list(ifaces.all_ifaces())) == 3
        assert ifaces.get_iface(
            f"{FOO1_IFACE_NAME}v0", InterfaceType.ETHERNET
        ).is_desired
        assert ifaces.get_iface(
            f"{FOO1_IFACE_NAME}v1", InterfaceType.ETHERNET
        ).is_desired

    def test_ignore_vf_when_pf_is_down(self):
        des_iface_infos = [gen_foo_iface_info(InterfaceType.ETHERNET)]
        des_iface_infos[0][Interface.NAME] = FOO1_IFACE_NAME
        des_iface_infos[0][Interface.STATE] = InterfaceState.DOWN
        des_iface_infos[0][Ethernet.CONFIG_SUBTREE] = {
            Ethernet.SRIOV_SUBTREE: {Ethernet.SRIOV.TOTAL_VFS: 2}
        }
        ifaces = Ifaces(des_iface_infos, [])
        assert len(list(ifaces.all_ifaces())) == 1
        assert ifaces.get_iface(
            FOO1_IFACE_NAME, InterfaceType.ETHERNET
        ).is_desired

    def test_ignore_vf_when_pf_is_absent(self):
        des_iface_infos = [gen_foo_iface_info(InterfaceType.ETHERNET)]
        des_iface_infos[0][Interface.NAME] = FOO1_IFACE_NAME
        des_iface_infos[0][Interface.STATE] = InterfaceState.ABSENT
        des_iface_infos[0][Ethernet.CONFIG_SUBTREE] = {
            Ethernet.SRIOV_SUBTREE: {Ethernet.SRIOV.TOTAL_VFS: 2}
        }
        ifaces = Ifaces(des_iface_infos, [])
        assert len(list(ifaces.all_ifaces())) == 1
        assert ifaces.get_iface(
            FOO1_IFACE_NAME, InterfaceType.ETHERNET
        ).is_desired



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import json
import os
import time

from libnmstate import __version__
from libnmstate.error import NmstateConflictError
from libnmstate.schema import Constants

from .testlib import assertlib
from .testlib import cmdlib
from .testlib.examplelib import example_state
from .testlib.examplelib import find_examples_dir
from .testlib.examplelib import load_example


SET_CMD = ["nmstatectl", "set"]
SHOW_CMD = ["nmstatectl", "show"]
CONFIRM_CMD = ["nmstatectl", "commit"]
ROLLBACK_CMD = ["nmstatectl", "rollback"]

LOOPBACK_JSON_CONFIG = """        {
            "name": "lo",
            "type": "unknown",
            "state": "up",
            "ipv4": {
                "enabled": true,
                "address": [
                    {
                        "ip": "127.0.0.1",
                        "prefix-length": 8
                    }
                ]
            },
            "ipv6": {
                "enabled": true,
                "address": [
                    {
                        "ip": "::1",
                        "prefix-length": 128
                    }
                ]
            },
            "mac-address": "00:00:00:00:00:00",
            "mtu": 65536
        }"""

LOOPBACK_YAML_CONFIG = """- name: lo
  type: unknown
  state: up
  ipv4:
    enabled: true
    address:
    - ip: 127.0.0.1
      prefix-length: 8
  ipv6:
    enabled: true
    address:
    - ip: ::1
      prefix-length: 128
  mac-address: 00:00:00:00:00:00
  mtu: 65536"""

ETH1_YAML_CONFIG = b"""interfaces:
- name: eth1
  state: up
  type: ethernet
  ipv4:
    address:
    - ip: 192.0.2.250
      prefix-length: 24
    enabled: true
  ipv6:
    enabled: false
  mtu: 1500
"""

EXAMPLES = find_examples_dir()
CONFIRMATION_INTERFACE = "eth1.101"
CONFIRMATION_CLEAN = "vlan101_eth1_absent.yml"
CONFIRMATION_TEST = "vlan101_eth1_up.yml"
CONFIRMATION_TEST_STATE = load_example(CONFIRMATION_TEST)
CONFIRMATION_SET = SET_CMD + [
    "--no-commit",
    os.path.join(EXAMPLES, CONFIRMATION_TEST),
]
CONFIRMATION_TIMEOUT = 5
CONFIRMATION_TIMOUT_COMMAND = SET_CMD + [
    "--no-commit",
    "--timeout",
    str(CONFIRMATION_TIMEOUT),
    os.path.join(EXAMPLES, CONFIRMATION_TEST),
]


def test_missing_operation():
    cmds = ["nmstatectl", "no-such-oper"]
    ret = cmdlib.exec_cmd(cmds)
    rc, out, err = ret

    assert rc == cmdlib.RC_FAIL2, cmdlib.format_exec_cmd_result(ret)
    assert "nmstatectl: error: invalid choice: 'no-such-oper'" in err


def test_show_command_with_json():
    ret = cmdlib.exec_cmd(SHOW_CMD + ["--json"])
    rc, out, err = ret

    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)
    assert LOOPBACK_JSON_CONFIG in out

    state = json.loads(out)
    assert len(state[Constants.INTERFACES]) > 1


def test_show_command_with_yaml_format():
    ret = cmdlib.exec_cmd(SHOW_CMD)
    rc, out, err = ret

    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)
    assert LOOPBACK_YAML_CONFIG in out


def test_show_command_json_only_lo():
    ret = cmdlib.exec_cmd(SHOW_CMD + ["--json", "lo"])
    rc, out, err = ret

    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)

    state = json.loads(out)
    assert len(state[Constants.INTERFACES]) == 1
    assert state[Constants.INTERFACES][0]["name"] == "lo"


def test_show_command_only_non_existing():
    ret = cmdlib.exec_cmd(SHOW_CMD + ["--json", "non_existing_interface"])
    rc, out, err = ret

    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)

    state = json.loads(out)
    assert len(state[Constants.INTERFACES]) == 0


def test_set_command_with_yaml_format():
    ret = cmdlib.exec_cmd(SET_CMD, stdin=ETH1_YAML_CONFIG)
    rc, out, err = ret

    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)


def test_set_command_with_two_states():
    examples = find_examples_dir()
    cmd = SET_CMD + [
        os.path.join(examples, "linuxbrige_eth1_up.yml"),
        os.path.join(examples, "linuxbrige_eth1_absent.yml"),
    ]
    ret = cmdlib.exec_cmd(cmd)
    rc = ret[0]

    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)
    assertlib.assert_absent("linux-br0")


def test_manual_confirmation(eth1_up):
    """ I can manually confirm a state. """

    with example_state(CONFIRMATION_CLEAN, CONFIRMATION_CLEAN):

        assert_command(CONFIRMATION_SET)
        assertlib.assert_state(CONFIRMATION_TEST_STATE)
        assert_command(CONFIRM_CMD)
        assertlib.assert_state(CONFIRMATION_TEST_STATE)


def test_manual_rollback(eth1_up):
    """ I can manually roll back a state. """

    with example_state(CONFIRMATION_CLEAN, CONFIRMATION_CLEAN) as clean_state:

        assert_command(CONFIRMATION_SET)
        assertlib.assert_state(CONFIRMATION_TEST_STATE)
        assert_command(ROLLBACK_CMD)
        assertlib.assert_state(clean_state)


def test_dual_change(eth1_up):
    """
    I cannot set a state without confirming/rolling back the state change.
    """

    with example_state(CONFIRMATION_CLEAN, CONFIRMATION_CLEAN) as clean_state:

        assert_command(CONFIRMATION_SET)
        assertlib.assert_state(CONFIRMATION_TEST_STATE)

        try:
            cmdlib.exec_cmd(CONFIRMATION_SET)
        except Exception as e:
            assert isinstance(e, NmstateConflictError)
        finally:
            assert_command(ROLLBACK_CMD)
            assertlib.assert_state(clean_state)


def test_automatic_rollback(eth1_up):
    """ If I do not confirm the state, it is automatically rolled back. """

    with example_state(CONFIRMATION_CLEAN, CONFIRMATION_CLEAN) as clean_state:

        assert_command(CONFIRMATION_TIMOUT_COMMAND)
        assertlib.assert_state(CONFIRMATION_TEST_STATE)

        time.sleep(CONFIRMATION_TIMEOUT)
        assertlib.assert_state(clean_state)


def test_version_argument():
    ret = cmdlib.exec_cmd(("nmstatectl", "--version"))
    rc, out, _ = ret
    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)
    assert out.rstrip() == __version__


def test_version_command():
    ret = cmdlib.exec_cmd(("nmstatectl", "version"))
    rc, out, _ = ret
    assert rc == cmdlib.RC_SUCCESS, cmdlib.format_exec_cmd_result(ret)
    assert out.rstrip() == __version__


def assert_command(cmd, expected_rc=cmdlib.RC_SUCCESS):
    ret = cmdlib.exec_cmd(cmd)
    returncode = ret[0]

    assert returncode == expected_rc, cmdlib.format_exec_cmd_result(ret)
    return ret



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate
from libnmstate.error import NmstateNotImplementedError
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import Route

from .testlib import statelib

IPV4_DNS_NAMESERVERS = ["8.8.8.8", "1.1.1.1"]
EXTRA_IPV4_DNS_NAMESERVER = "9.9.9.9"
IPV6_DNS_NAMESERVERS = ["2001:4860:4860::8888", "2606:4700:4700::1111"]
EXTRA_IPV6_DNS_NAMESERVER = "2620:fe::9"
EXAMPLE_SEARCHES = ["example.org", "example.com"]

parametrize_ip_ver = pytest.mark.parametrize(
    "dns_config",
    [
        ({DNS.SERVER: IPV4_DNS_NAMESERVERS, DNS.SEARCH: EXAMPLE_SEARCHES}),
        ({DNS.SERVER: IPV6_DNS_NAMESERVERS, DNS.SEARCH: EXAMPLE_SEARCHES}),
    ],
    ids=["ipv4", "ipv6"],
)

DUMMY0 = "dummy0"


@pytest.fixture(scope="function", autouse=True)
def dns_test_env(eth1_up, eth2_up):
    yield
    # Remove DNS config as it be saved in eth1 or eth2 which might trigger
    # failure when bring eth1/eth2 down.
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        DNS.KEY: {DNS.CONFIG: {DNS.SERVER: [], DNS.SEARCH: []}},
    }
    libnmstate.apply(desired_state)


@pytest.mark.tier1
@parametrize_ip_ver
def test_dns_edit_nameserver_with_static_gateway(dns_config):
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


def test_dns_edit_ipv4_nameserver_before_ipv6():
    dns_config = {
        DNS.SERVER: [IPV4_DNS_NAMESERVERS[0], IPV6_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.mark.tier1
def test_dns_edit_ipv6_nameserver_before_ipv4():
    dns_config = {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.mark.tier1
@pytest.mark.parametrize(
    "dns_servers",
    [
        (IPV4_DNS_NAMESERVERS + [EXTRA_IPV4_DNS_NAMESERVER]),
        (IPV6_DNS_NAMESERVERS + [EXTRA_IPV6_DNS_NAMESERVER]),
        (IPV4_DNS_NAMESERVERS + [EXTRA_IPV6_DNS_NAMESERVER]),
        (IPV6_DNS_NAMESERVERS + [EXTRA_IPV4_DNS_NAMESERVER]),
        pytest.param(
            (
                [
                    IPV4_DNS_NAMESERVERS[0],
                    EXTRA_IPV6_DNS_NAMESERVER,
                    IPV4_DNS_NAMESERVERS[1],
                ]
            ),
            marks=pytest.mark.xfail(
                reason="Not supported",
                raises=NmstateNotImplementedError,
                strict=True,
            ),
        ),
        pytest.param(
            (
                [
                    IPV6_DNS_NAMESERVERS[0],
                    EXTRA_IPV4_DNS_NAMESERVER,
                    IPV6_DNS_NAMESERVERS[1],
                ]
            ),
            marks=pytest.mark.xfail(
                reason="Not supported",
                raises=NmstateNotImplementedError,
                strict=True,
            ),
        ),
        (IPV4_DNS_NAMESERVERS + IPV6_DNS_NAMESERVERS),
        (IPV6_DNS_NAMESERVERS + IPV4_DNS_NAMESERVERS),
    ],
    ids=[
        "3ipv4",
        "3ipv6",
        "2ipv4+ipv6",
        "2ipv6+ipv4",
        "ipv4+ipv6+ipv4",
        "ipv6+ipv4+ipv6",
        "2ipv4+2ipv6",
        "2ipv6+2ipv4",
    ],
)
def test_dns_edit_3_more_nameservers(dns_servers):
    dns_config = {
        DNS.SERVER: dns_servers,
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)
    current_state = libnmstate.show()
    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.mark.tier1
@pytest.mark.parametrize(
    "empty_dns_config",
    [{DNS.SERVER: []}, {DNS.SEARCH: []}, {DNS.SERVER: [], DNS.SEARCH: []}, {}],
    ids=[
        "empty_server",
        "empty_search",
        "empty_server_and_search",
        "empty_dict",
    ],
)
def test_remove_dns_config(empty_dns_config):
    dns_config = {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)

    libnmstate.apply(
        {Interface.KEY: [], DNS.KEY: {DNS.CONFIG: empty_dns_config}}
    )
    current_state = libnmstate.show()
    assert {} == current_state[DNS.KEY][DNS.CONFIG]


@pytest.fixture
def dummy0_up():
    dummy_iface_state = {
        Interface.NAME: DUMMY0,
        Interface.TYPE: InterfaceType.DUMMY,
        Interface.STATE: InterfaceState.UP,
    }
    libnmstate.apply({Interface.KEY: [dummy_iface_state]})
    yield dummy_iface_state
    dummy_iface_state[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply({Interface.KEY: [dummy_iface_state]})


def test_preserve_dns_config(dummy0_up):
    dns_config = {
        DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
        DNS.SEARCH: [],
    }
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {DNS.CONFIG: dns_config},
    }
    libnmstate.apply(desired_state)

    # Add new dummy interface with default gateway, nmstate should
    # preserve the existing DNS configure as interface holding DNS
    # configuration is not changed and DNS configure is still the same.
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: DUMMY0,
                    Interface.TYPE: InterfaceType.DUMMY,
                    Interface.IPV4: {
                        InterfaceIPv4.ADDRESS: [
                            {
                                InterfaceIPv4.ADDRESS_IP: "192.0.2.250",
                                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                            }
                        ],
                        InterfaceIPv4.ENABLED: True,
                    },
                    Interface.IPV6: {
                        InterfaceIPv6.ADDRESS: [
                            {
                                InterfaceIPv6.ADDRESS_IP: "2001:db8:f::1",
                                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                            }
                        ],
                        InterfaceIPv6.ENABLED: True,
                    },
                }
            ],
            Route.KEY: {
                Route.CONFIG: [
                    {
                        Route.DESTINATION: "0.0.0.0/0",
                        Route.NEXT_HOP_ADDRESS: "192.0.2.2",
                        Route.NEXT_HOP_INTERFACE: DUMMY0,
                    },
                    {
                        Route.DESTINATION: "::/0",
                        Route.NEXT_HOP_ADDRESS: "2001:db8:f::2",
                        Route.NEXT_HOP_INTERFACE: DUMMY0,
                    },
                ]
            },
        }
    )

    current_state = libnmstate.show()

    assert dns_config == current_state[DNS.KEY][DNS.CONFIG]


@pytest.fixture
def setup_ipv4_ipv6_name_server():
    desired_state = {
        Interface.KEY: _get_test_iface_states(),
        Route.KEY: {Route.CONFIG: _gen_default_gateway_route()},
        DNS.KEY: {
            DNS.CONFIG: {
                DNS.SERVER: [IPV6_DNS_NAMESERVERS[0], IPV4_DNS_NAMESERVERS[0]],
                DNS.SEARCH: [],
            }
        },
    }
    libnmstate.apply(desired_state)
    yield desired_state


def test_preserve_dns_config_with_empty_state(setup_ipv4_ipv6_name_server):
    old_state = setup_ipv4_ipv6_name_server

    libnmstate.apply({Interface.KEY: []})
    current_state = libnmstate.show()

    assert old_state[DNS.KEY][DNS.CONFIG] == current_state[DNS.KEY][DNS.CONFIG]


def _get_test_iface_states():
    return [
        {
            Interface.NAME: "eth1",
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Interface.IPV4: {
                InterfaceIPv4.ADDRESS: [
                    {
                        InterfaceIPv4.ADDRESS_IP: "192.0.2.251",
                        InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                    }
                ],
                InterfaceIPv4.DHCP: False,
                InterfaceIPv4.ENABLED: True,
            },
            Interface.IPV6: {
                InterfaceIPv6.ADDRESS: [
                    {
                        InterfaceIPv6.ADDRESS_IP: "2001:db8:1::1",
                        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                    }
                ],
                InterfaceIPv6.DHCP: False,
                InterfaceIPv6.AUTOCONF: False,
                InterfaceIPv6.ENABLED: True,
            },
        },
        {
            Interface.NAME: "eth2",
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Interface.IPV4: {
                InterfaceIPv4.ADDRESS: [
                    {
                        InterfaceIPv4.ADDRESS_IP: "198.51.100.1",
                        InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                    }
                ],
                InterfaceIPv4.DHCP: False,
                InterfaceIPv4.ENABLED: True,
            },
            Interface.IPV6: {
                InterfaceIPv6.ADDRESS: [
                    {
                        InterfaceIPv6.ADDRESS_IP: "2001:db8:2::1",
                        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                    }
                ],
                InterfaceIPv6.DHCP: False,
                InterfaceIPv6.AUTOCONF: False,
                InterfaceIPv6.ENABLED: True,
            },
        },
    ]


def _gen_default_gateway_route():
    return [
        {
            Route.DESTINATION: "0.0.0.0/0",
            Route.METRIC: 200,
            Route.NEXT_HOP_ADDRESS: "192.0.2.1",
            Route.NEXT_HOP_INTERFACE: "eth1",
        },
        {
            Route.DESTINATION: "::/0",
            Route.METRIC: 201,
            Route.NEXT_HOP_ADDRESS: "2001:db8:2::f",
            Route.NEXT_HOP_INTERFACE: "eth1",
        },
    ]


def test_show_saved_config_with_dns_interface_down(
    setup_ipv4_ipv6_name_server,
):
    running_dns_config = libnmstate.show()[DNS.KEY][DNS.CONFIG]
    libnmstate.apply(
        {
            # Below are requred as we bring down the DNS interface,
            # nmstate cannot find new interface for the old DNS config.
            DNS.KEY: {DNS.CONFIG: {}},
            Interface.KEY: [
                {
                    Interface.NAME: "eth1",
                    Interface.STATE: InterfaceState.DOWN,
                }
            ],
        }
    )
    saved_iface_state = statelib.show_saved_config_only(("eth1",))[
        Interface.KEY
    ][0]
    saved_dns_config = libnmstate.show_saved_config()[DNS.KEY][DNS.CONFIG]

    assert saved_iface_state[Interface.STATE] == InterfaceState.UP
    assert saved_dns_config == running_dns_config
